# coding: utf-8

"""
module to define a tomography z-series acquisition (made by bliss)
"""

from __future__ import annotations

import pint

import h5py
import numpy
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from silx.utils.proxy import docstring

from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.utils.io import deprecated
from nxtomomill.models.utils import _get_title_dataset

from .baseacquisition import BaseAcquisition, EntryReader
from .standardacquisition import StandardAcquisition

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass

_ureg = pint.get_application_registry()


__all__ = [
    "is_z_series_sequence",
    "is_pcotomo_frm_titles",
    "is_z_series_frm_translation_z",
    "ZSeriesBaseAcquisition",
]


def is_z_series_sequence(
    entry: h5py.Group, configuration: TomoHDF5Config, fallback_on_title: bool = True
) -> bool:
    """
    Check if the provided h5py.Group must be consider as an "initialization" of a sequence. And if so then if it is a back-and-forth sequence.
    It will first check for value contained in technique/scan_category else for the title name (legacy)
    """
    # check 'technique/scan_category' first
    # warning: checking the 'scan_category' is more for legacy. Today the z-series sequence has been replace by a set of standard acquisition sequences.
    scan_category = entry.get("technique/scan_category", None)
    if scan_category is not None:
        category_name = h5py_read_dataset(scan_category)
    elif fallback_on_title:
        # else fallback on title check
        category_name = _get_title_dataset(
            entry=entry, title_paths=BaseAcquisition.TITLE_PATHS
        )
    else:
        category_name = None

    if category_name is None:
        return False

    for init_title in configuration.zseries_init_titles:
        if category_name.startswith(init_title):
            return True
    return False


@deprecated(
    since_version="1.2",
    reason="rename pcotomo to multitomo",
    replacement="nxtomomill.converter.hdf5.acquisition.acquisitionConstructor.is_multitomo_frm_titles",
)
def is_pcotomo_frm_titles(entry: h5py.Group, configuration: TomoHDF5Config) -> bool:
    from ..acquisitionConstructor import is_multitomo_sequence

    return is_multitomo_sequence(entry=entry, configuration=configuration)


def is_multitomo_frm_titles(entry: h5py.Group, configuration: TomoHDF5Config) -> bool:
    """
    if the provided h5py.Group must be consider as an "initialization" entry/scan of a multitomo acquisition
    """
    title = _get_title_dataset(entry=entry, title_paths=BaseAcquisition.TITLE_PATHS)
    if title is None:
        return False

    for multitomo_init_title in configuration.multitomo_init_titles:
        if title.startswith(multitomo_init_title):
            return True
    return False


def is_z_series_frm_translation_z(projection_urls, configuration: TomoHDF5Config):
    """
    :param projection_urls: list of DataUrl pointing to projection
                                     nodes.
    :return: True if the set of projections should be considered as a z-series (according to version 1 of z-series)
    """
    z_values = set()
    for url in projection_urls:
        with EntryReader(url) as entry:
            z_values_tmp = BaseAcquisition.get_translation_z_frm(
                entry, n_frame=None, configuration=configuration
            )
            if z_values_tmp is not None:
                magnitudes = numpy.asarray(z_values_tmp.to(_ureg.meter).magnitude)
                if magnitudes.ndim == 0:
                    z_values.add(magnitudes.item())
                else:
                    z_values.update(magnitudes.flatten().tolist())
    return len(z_values) > 1


class ZSeriesBaseAcquisition(BaseAcquisition):
    """
    A 'z series acquisition' is considered as a series of _StandardAcquisition.
    Registered scan can be split according to translation_z value.

    At the moment there is three version of z-series:

    #. **version 1**: each z is part of the same sequence. bliss .h5 will look like:
        * 1.1    tomo:zserie                                          -> define the beginning of the sequence
        * 2.1    reference images1 (flats)                            -> start of the first z level
        * 3.1    projections 1 -7000 (flats)
        * 4.1    static images (alignment / return projections)
        * 5.1    reference images1 (flats)
        * 6.1    dark images
        * 7.1    reference images1 (flats)                            -> start of the second z level. using `get_z` to know the different levels
        * 8.1    projections 1 -7000 (flats)
        * ...
        in this case an instance of ZSeriesBaseAcquisition will create N NXtomo (one nxtomo per sequence)

    #. **version 2**: each z is part of a new sequence. So each sequence will instantiate ZSeriesBaseAcquisition and each with a single z.
        * 1.1    tomo:zserie                                          -> define the beginning of the sequence
        * 2.1    reference images1 (flats)                            -> start of the first z level
        * 3.1    projections 1 -7000 (flats)
        * 4.1    static images (alignment / return projections)
        * 5.1    reference images1 (flats)
        * 6.1    dark images
        * 7.1    tomo:zserie                                          -> define the beginning of the sequence
        * 8.1    reference images1 (flats)                            -> start of the second z level. using `get_z` to know the different levels
        * 9.1    projections 1 -7000 (flats)
        * ...
        in this case an instance of ZSeriesBaseAcquisition will create one NXtomo (one NXtomo per sequence)


    #. **version 3**: same as version 2 but dark / flat can only be done in a at the beginning or at the end of the series. And we want to copy those.
    To keep compatibility and design this part is done in post-processing.
        in this case an instance of ZSeriesBaseAcquisition will also create one NXtomo (one NXtomo per sequence)

    The goal of this class is mostly to handle the version 1.
    For version 2 and 3 it will be instantiated but `_acquisition` will contain a single acquisition.
    But to manipulate the series in the case of version 2 and especially version 3 the converter will group them inside `_z_series_v2_v3`
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._acquisitions: tuple[float, StandardAcquisition] = {}
        """key is z value and value is StandardAcquisition"""
        (
            self._dark_at_start,
            self._dark_at_end,
            self._flat_at_start,
            self._flat_at_end,
        ) = self.get_dark_flat_pos_info()

    def get_dark_flat_pos_info(self) -> tuple:
        #  on latest z-series dark can be saved only at the beginning / end
        # so we need to copy those in post processing
        if self.root_url is None:
            return None, None, None, None
        with EntryReader(self.root_url) as entry:
            scan_flags_grp = self._get_scan_flags(entry_node=entry) or {}
            dark_at_start_dataset = scan_flags_grp.get("dark_images_at_start", None)
            if dark_at_start_dataset is not None:
                dark_at_start_dataset = h5py_read_dataset(dark_at_start_dataset)
            dark_at_end_dataset = scan_flags_grp.get("dark_images_at_end", None)
            if dark_at_end_dataset is not None:
                dark_at_end_dataset = h5py_read_dataset(dark_at_end_dataset)
            flat_at_start_dataset = scan_flags_grp.get(
                "ref_images_at_start", scan_flags_grp.get("flat_images_at_start", None)
            )
            if flat_at_start_dataset is not None:
                flat_at_start_dataset = h5py_read_dataset(flat_at_start_dataset)
            flat_at_end_dataset = scan_flags_grp.get(
                "ref_images_at_end", scan_flags_grp.get("flat_images_at_end", None)
            )
            if flat_at_end_dataset is not None:
                flat_at_end_dataset = h5py_read_dataset(flat_at_end_dataset)

        return (
            dark_at_start_dataset,
            dark_at_end_dataset,
            flat_at_start_dataset,
            flat_at_end_dataset,
        )

    def get_expected_nx_tomo(self):
        return 1

    def get_standard_sub_acquisitions(self) -> tuple:
        """
        Return the tuple of all :class:`.StandardAcquisition` composing
        _acquisitions
        """
        return tuple(self._acquisitions.values())

    def get_z(self, entry):
        if not isinstance(entry, h5py.Group):
            raise TypeError("entry: expected h5py.Group")
        z_array: pint.Quantity | None = self._get_translation_z(entry, n_frame=None)
        if z_array is None:
            raise ValueError(f"No z found for scan {entry.name}")
        z_array = z_array.to_base_units().magnitude
        if isinstance(z_array, numpy.ndarray):
            z_array = set(z_array)
        else:
            z_array = set((z_array,))

        # might need an epsilon here ?
        if len(z_array) > 1:
            raise ValueError(f"More than one value of z found for {entry.name}")
        else:
            return z_array.pop()

    @docstring(BaseAcquisition.register_step)
    def register_step(
        self, url: DataUrl, entry_type, copy_frames: bool = False
    ) -> None:
        """

        :param url:
        """
        with EntryReader(url) as entry:
            z = self.get_z(entry)
        if z not in self._acquisitions:
            new_acquisition = StandardAcquisition(
                root_url=url,
                configuration=self.configuration,
                detector_sel_callback=self._detector_sel_callback,
                start_index=self.start_index + len(self._acquisitions),
                parent=self,
            )
            new_acquisition._dark_at_start = self._dark_at_start
            new_acquisition._flat_at_start = self._flat_at_start
            new_acquisition._dark_at_end = self._dark_at_end
            new_acquisition._flat_at_end = self._flat_at_end
            self._acquisitions[z] = new_acquisition
        self._acquisitions[z].register_step(
            url=url, entry_type=entry_type, copy_frames=copy_frames
        )



"""Scan dedicated for bliss fluo-data format - based on h5 files generated by ewoksfluo"""

from __future__ import annotations

import logging
import pint

import numpy

from dataclasses import dataclass, field
import numpy as np
from numpy.typing import NDArray, DTypeLike

import h5py

_ureg = pint.get_application_registry()
_logger = logging.getLogger(__name__)

try:
    import tifffile  # noqa #F401 needed for later possible lazy loading
except ImportError:
    has_tifffile = False
else:
    has_tifffile = True

__all__ = ["BlissFluoTomoScanBase", "BlissFluoTomoScan3D", "BlissFluoTomoScan2D"]


@dataclass
class BlissFluoTomoScanBase:
    """Base class to read XRF data fitted with ewoksfluo."""

    ewoksfluo_filename: str
    detectors: tuple = ()
    dtype: DTypeLike = numpy.float32
    verbose: bool = False
    angles_deg: list[dtype] | None = None
    lines: dict[str, list[str]] = field(default_factory=dict)
    pixel_size: float | None = None
    energy: float | None = None

    @property
    def rot_angles_deg(self) -> NDArray:
        return self.angles_deg

    @property
    def rot_angles_rad(self) -> NDArray:
        return self.rot_angles_deg.to(_ureg.rad)


@dataclass
class BlissFluoTomoScan3D(BlissFluoTomoScanBase):
    """Base class to read XRF data fitted with ewoksfluo."""

    ewoksfluo_filename: str
    detectors: tuple = ()
    dtype: DTypeLike = numpy.float32
    verbose: bool = False
    angles_deg: list[dtype] | None = None
    lines: list[str] = field(default_factory=list)
    pixel_size: float | None = None
    energy: float | None = None

    def _check_entry(self, entry, f):
        data_shape = f[
            f"{entry}/grid/{self.detectors[0]}/results/massfractions/{self.lines[0]}"
        ][()].shape
        flag = self.slow_npoints * self.fast_npoints == data_shape[0] * data_shape[1]
        if not flag:
            msg = f"Entry {entry} has unexpected data shape {data_shape}, expected {(self.slow_npoints, self.fast_npoints)}. It has been ignored (i.e. not converted)."
            _logger.warning(msg)
        return flag

    def __post_init__(self):
        try:
            with h5py.File(self.ewoksfluo_filename, "r") as f:
                self.entries = sorted(list(f.keys()), key=lambda x: float(x))
                entry0 = self.entries[0]

                # Get detector names
                if len(self.detectors) == 0:
                    self.detectors = list(f[f"{entry0}/grid"].keys())

                # Get fitted line names
                detector0 = self.detectors[0]
                self.lines = []
                for name, ds in f[
                    f"{entry0}/grid/{detector0}/results/massfractions"
                ].items():
                    if ds.ndim == 2:
                        self.lines.append(name)

                # Get projections dimensions
                self.slow_npoints = float(
                    f[f"{entry0}/instrument/fscan_parameters/slow_npoints"][()]
                )
                self.fast_npoints = float(
                    f[f"{entry0}/instrument/fscan_parameters/fast_npoints"][()]
                )

                # Filter out interrupted scans:
                self.entries = [
                    entry for entry in self.entries if self._check_entry(entry, f)
                ]

                # Get rotation angles
                angles_tmp = []
                for entry in self.entries:
                    angles_tmp.append(f[f"{entry}/instrument/positioners/somega"][()])
                self.angles_deg = np.array(angles_tmp, dtype=self.dtype) * _ureg.deg

                # Get pixel size
                fast_pixel_size = float(
                    f[f"{entry0}/instrument/fscan_parameters/fast_step_size"][()]
                )
                slow_pixel_size = float(
                    f[f"{entry0}/instrument/fscan_parameters/slow_step_size"][()]
                )
                if slow_pixel_size == fast_pixel_size:
                    self.pixel_size = slow_pixel_size * _ureg.um
                else:
                    msg = f"Found slow_pixel_size ({slow_pixel_size}) different from fast_pixel_size ({fast_pixel_size})"
                    _logger.info(msg)
                    raise ValueError(msg)

                # Get energy
                self.energy = (
                    float(
                        f[
                            f"{entry0}/instrument/metadata/InstrumentMonochromator_energy"
                        ][()]
                    )
                    * _ureg.keV
                )

        except FileNotFoundError:
            self.entries = None
            _logger.info(
                f"Detected {len(self.entries)} projections in {self.ewoksfluo_filename}."
            )

    def load_data(self, det, line):
        if self.entries is None:
            _logger.info(f"No entry was detected in {self.ewoksfluo_filename}.")
        else:
            with h5py.File(self.ewoksfluo_filename, "r") as f:
                projs_tmp = []
                angles_tmp = []
                for entry in self.entries:
                    data = f[f"{entry}/grid/{det}/results/massfractions/{line}"][()]
                    projs_tmp.append(numpy.nan_to_num(data).astype(self.dtype))
                    angles_tmp.append(f[f"{entry}/instrument/positioners/somega"][()])

            projs_tmp = np.array(projs_tmp)
            angles_tmp = np.array(angles_tmp).astype(self.dtype)
            if np.allclose(angles_tmp * _ureg.deg, self.angles_deg):
                return np.ascontiguousarray(projs_tmp)
            else:
                msg = f"Angles found for {det}/{line} do not match the angles found at the class level."
                _logger.info(msg)
                raise ValueError(msg)


@dataclass
class BlissFluoTomoScan2D(BlissFluoTomoScanBase):
    """Base class to read XRF data fitted with ewoksfluo."""

    ewoksfluo_filename: str
    detectors: tuple = ()
    dtype: DTypeLike = numpy.float32
    verbose: bool = False
    angles_deg: list[dtype] | None = None
    lines: list[str] = field(default_factory=list)
    pixel_size: float | None = None
    energy: float | None = None



"""python API to convert from several format to `NXtomo <https://manual.nexusformat.org/classes/applications/NXtomo.html>`_"""

from .dxfile.dxfileconverter import from_dx_to_nx  # noqa F401
from .edf.edfconverter import EDFFileKeys  # noqa F401
from .edf.edfconverter import edf_to_nx  # noqa F401
from .edf.edfconverter import from_edf_to_nx  # noqa F401
from .hdf5.hdf5converter import from_h5_to_nx  # noqa F401
from .hdf5.hdf5converter import get_bliss_tomo_entries  # noqa F401
from .fluo.fluoconverter import from_fluo_to_nx  # noqa F401
from .fluo.fluoconverter import from_blissfluo_to_nx  # noqa F401



import pytest
import os
import numpy
from tomoscan.io import HDF5File, get_swmr_mode
from nxtomomill.converter.hdf5.acquisition.blisstomoconfig import TomoConfig


def test_tomo_config(tmp_path):
    """
    test a configuration from bliss can be read from nxtomomill
    """
    file_path = os.path.join(tmp_path, "test_bliss_tomo_config.hdf5")
    with HDF5File(file_path, mode="w") as h5s:
        tomo_config_group = h5s.create_group("/0/test/instrument/tomoconfig")
        tomo_config_group["rotation"] = ("my_rot", "rot")
        tomo_config_group["sample_u"] = numpy.array(
            [
                "alias",
                "motor_name",
            ],
            dtype=object,
        )
        tomo_config_group["sample_v"] = numpy.array(
            [
                "sample_v_motor",
            ],
            dtype=object,
        )
        tomo_config_group["sample_x"] = [
            "sample_x_motor",
        ]
        tomo_config_group["sample_y"] = numpy.array(
            [
                "sample_y_motor",
            ],
            dtype=object,
        )
        tomo_config_group["detector"] = ["alias", "frelon"]
        tomo_config_group["translation_y"] = numpy.array(
            [
                "ty",
            ],
            dtype=object,
        )
        tomo_config_group["translation_z"] = numpy.array(
            [
                "tz",
            ],
            dtype=object,
        )
        tomo_config_group["rotation_is_clockwise"] = True

    with pytest.raises(TypeError):
        TomoConfig.from_technique_group(file_path)

    with pytest.raises(KeyError):
        TomoConfig.from_technique_group(h5s)

    with HDF5File(file_path, mode="r", swmr=get_swmr_mode()) as h5s:
        tomo_config = TomoConfig.from_technique_group(h5s["/0/test/instrument"])

    numpy.testing.assert_array_equal(
        tomo_config.rotation, numpy.array(["my_rot", "rot"], dtype=object)
    )
    numpy.testing.assert_array_equal(
        tomo_config.sample_u, numpy.array(["alias", "motor_name"], dtype=object)
    )
    numpy.testing.assert_array_equal(
        tomo_config.sample_v,
        numpy.array(
            [
                "sample_v_motor",
            ],
            dtype=object,
        ),
    )
    numpy.testing.assert_array_equal(
        tomo_config.sample_x,
        numpy.array(
            [
                "sample_x_motor",
            ],
            dtype=object,
        ),
    )
    numpy.testing.assert_array_equal(
        tomo_config.sample_y,
        numpy.array(
            [
                "sample_y_motor",
            ],
            dtype=object,
        ),
    )
    numpy.testing.assert_array_equal(
        tomo_config.tomo_detector,
        numpy.array(
            [
                "alias",
                "frelon",
            ],
            dtype=object,
        ),
    )
    numpy.testing.assert_array_equal(
        tomo_config.translation_y,
        numpy.array(
            [
                "ty",
            ],
            dtype=object,
        ),
    )
    numpy.testing.assert_array_equal(
        tomo_config.translation_z,
        numpy.array(
            [
                "tz",
            ],
            dtype=object,
        ),
    )

    assert tomo_config.rotation_is_clockwise is True



from __future__ import annotations


from pydantic import BaseModel, ConfigDict, Field, field_validator
from nxtomomill.models.utils import filter_str_def


class InstrumentSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    instrument_name: str | None = Field(
        default=None, description="Name of the instrument"
    )

    @field_validator(
        "instrument_name",
        mode="plain",
    )
    @classmethod
    def cast_instrument_name(cls, value: str | None) -> str | None:
        res = filter_str_def(value)
        return res



# coding: utf-8

"""
utils functions for io
"""

from __future__ import annotations

import logging
import re
import h5py
import numpy

from silx.utils.enum import Enum as _Enum
from silx.io.utils import h5py_read_dataset

_logger = logging.getLogger(__name__)


__all__ = [
    "remove_parenthesis_or_brackets",
    "filter_str_def",
    "convert_str_to_tuple",
    "convert_str_to_bool",
    "is_url_path",
    "PathType",
]


def remove_parenthesis_or_brackets(input_str):
    if (
        input_str.startswith("(")
        and input_str.endswith(")")
        or input_str.startswith("[")
        and input_str.endswith("]")
    ):
        input_str = input_str[1:-1]
    return input_str


def filter_str_def(elmt):
    if elmt is None:
        return None
    assert isinstance(elmt, str)
    elmt = elmt.lstrip(" ").rstrip(" ")
    for character in ("'", '"'):
        if elmt.startswith(character) and elmt.endswith(character):
            elmt = elmt[1:-1]
    return elmt


def convert_str_to_tuple(input_str: str, none_if_empty: bool = False) -> tuple | None:
    """
    :param input_str: string to convert
    :param none_if_empty: if true and the conversion is an empty tuple
                               return None instead of an empty tuple
    """
    if isinstance(input_str, (list, set)):
        input_str = tuple(input_str)
    if isinstance(input_str, tuple):
        return input_str
    if input_str is None:
        input_str = ""
    if not isinstance(input_str, str):
        raise TypeError(
            f"input_str should be a string not {type(input_str)}, {input_str}"
        )
    input_str = input_str.lstrip(" ").rstrip(" ")
    input_str = remove_parenthesis_or_brackets(input_str)

    elmts = input_str.split(",")
    elmts = [filter_str_def(elmt) for elmt in elmts]
    rm_empty_str = lambda a: a != ""
    elmts = list(filter(rm_empty_str, elmts))
    if none_if_empty and len(elmts) == 0:
        return None
    else:
        return tuple(elmts)


def convert_str_to_bool(value: str | bool | numpy.bool_):
    if isinstance(value, (bool, numpy.bool_)):
        return bool(value)
    elif isinstance(value, str):
        if value not in ("False", "True", "1", "0"):
            raise ValueError("value should be 'True' or 'False'")
        return value in ("True", "1")
    else:
        raise TypeError(f"value should be a string or a boolean. Got {type(value)}")


def is_url_path(url_str: str) -> bool:
    """
    :param url_str: url as a string
    :return: True if the provided string fit DataUrl pattern
        [scheme]:://[file_path]?[data_path]
    """
    pattern_str_seq = "[a-zA-Z0-9]*"
    url_path_pattern = rf"{pattern_str_seq}\:\/\/{pattern_str_seq}"
    pattern = re.compile(url_path_pattern)
    return bool(re.match(pattern, url_str))


def _get_title_dataset(entry: h5py.Group, title_paths: tuple[str]):
    for title_path in title_paths:
        if title_path in entry:
            return h5py_read_dataset(entry[title_path])
    return None


class PathType(_Enum):
    ABSOLUTE = "absolute"
    RELATIVE = "relative"



# coding: utf-8

from __future__ import annotations

import os
import shutil
import tempfile
import unittest

import h5py
import numpy.random

from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset

from tomoscan.io import HDF5File, get_swmr_mode
from tomoscan.esrf.mock import MockNXtomo as MockNXtomo
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan

from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.utils import add_dark_flat_nx_file, change_image_key_control


class MockNXtomoWithElecCurrent(MockNXtomo):
    def __init__(
        self,
        scan_path,
        ini_dark: numpy.array | None,
        ini_flats: numpy.array | None,
        final_flats: numpy.array | None,
        dim: int,
        n_proj: int,
        count_time: numpy.array | None = None,
        machine_current: numpy.array | None = None,
    ):
        assert ini_dark is None or ini_dark.ndim == 3, "ini_dark should be a 3d array"
        assert ini_flats is None or ini_flats.ndim == 3, "ini_dark should be a 3d array"
        assert (
            final_flats is None or final_flats.ndim == 3
        ), "ini_dark should be a 3d array"
        self._ini_darks = ini_dark
        self._ini_flats = ini_flats
        self._final_flats = final_flats
        self._count_time = count_time
        super().__init__(
            scan_path=scan_path,
            dim=dim,
            create_ini_dark=ini_dark is not None,
            create_ini_flat=ini_flats is not None,
            create_final_flat=final_flats is not None,
            n_ini_proj=n_proj,
            n_proj=n_proj,
        )

        # append count_time and machine_current to the HDF5 file
        with HDF5File(self.scan_master_file, "a") as h5_file:
            entry_one = h5_file.require_group(self.scan_entry)
            if machine_current is not None:
                monitor_grp = entry_one.require_group("control")
                monitor_grp["data"] = machine_current
            # rewrite count_time
            if count_time is not None:
                instrument_grp = entry_one.require_group("instrument")
                detector_grp = instrument_grp.require_group("detector")
                if "count_time" in detector_grp:
                    del detector_grp["count_time"]
                detector_grp["count_time"] = count_time

    def add_initial_dark(self):
        for frame in self._ini_darks:
            self._append_frame(
                data_=frame.reshape(1, frame.shape[0], frame.shape[1]),
                rotation_angle=self.rotation_angle[-1],
                image_key=ImageKey.DARK_FIELD.value,
                image_key_control=ImageKey.DARK_FIELD.value,
                diode_data=None,
            )

    def add_initial_flat(self):
        for frame in self._ini_flats:
            self._append_frame(
                data_=frame.reshape(1, frame.shape[0], frame.shape[1]),
                rotation_angle=self.rotation_angle[-1],
                image_key=ImageKey.FLAT_FIELD.value,
                image_key_control=ImageKey.FLAT_FIELD.value,
                diode_data=None,
            )

    def add_final_flat(self):
        for frame in self._final_flats:
            self._append_frame(
                data_=frame.reshape(1, frame.shape[0], frame.shape[1]),
                rotation_angle=self.rotation_angle[-1],
                image_key=ImageKey.FLAT_FIELD.value,
                image_key_control=ImageKey.FLAT_FIELD.value,
                diode_data=None,
            )


class BaseTestAddDarkAndFlats(unittest.TestCase):
    """
    Unit test on nxtomomill.utils.add_dark_flat_nx_file function
    """

    def setUp(self) -> None:
        self.tmpdir = tempfile.mkdtemp()
        simple_nx_path = os.path.join(self.tmpdir, "simple_case")
        self.dim = 55
        self.nproj = 20
        self._simple_nx = MockNXtomoWithElecCurrent(
            scan_path=simple_nx_path,
            n_proj=self.nproj,
            ini_dark=None,
            ini_flats=None,
            final_flats=None,
            dim=self.dim,
            machine_current=numpy.zeros(self.nproj),
        ).scan
        with HDF5File(
            self._simple_nx.master_file, mode="r", swmr=get_swmr_mode()
        ) as h5s:
            data_path = "/".join(
                (self._simple_nx.entry, "instrument", "detector", "data")
            )
            self._raw_data = h5py_read_dataset(h5s[data_path])
        nx_with_vds_path = os.path.join(self.tmpdir, "case_with_vds")
        self._nx_with_virtual_dataset = MockNXtomoWithElecCurrent(
            scan_path=nx_with_vds_path,
            n_proj=0,
            ini_dark=None,
            ini_flats=None,
            final_flats=None,
            dim=self.dim,
            machine_current=numpy.zeros(self.nproj),
        ).scan
        self._create_vds(
            source_file=self._simple_nx.master_file,
            source_data_path=self._simple_nx.entry,
            target_file=self._nx_with_virtual_dataset.master_file,
            target_data_path=self._nx_with_virtual_dataset.entry,
            copy_other_data=True,
        )
        self._patch_nxtomo_flags(self._nx_with_virtual_dataset)

        nx_with_vds_path_and_links = os.path.join(
            self.tmpdir, "case_with_vds_and_links"
        )
        self._nx_with_virtual_dataset_with_link = MockNXtomoWithElecCurrent(
            scan_path=nx_with_vds_path_and_links,
            n_proj=0,
            ini_dark=None,
            ini_flats=None,
            final_flats=None,
            dim=self.dim,
            machine_current=numpy.zeros(self.nproj),
        ).scan
        self._create_vds(
            source_file=self._simple_nx.master_file,
            source_data_path=self._simple_nx.entry,
            target_file=self._nx_with_virtual_dataset_with_link.master_file,
            target_data_path=self._nx_with_virtual_dataset_with_link.entry,
            copy_other_data=True,
        )
        self._patch_nxtomo_flags(self._nx_with_virtual_dataset_with_link)

        # create dark
        self.start_dark = (
            numpy.random.random((self.dim * self.dim))
            .reshape(1, self.dim, self.dim)
            .astype("f")
        )
        self.start_dark_file = os.path.join(self.tmpdir, "dark.hdf5")
        self.start_dark_entry = "data"
        self.start_dark_url = self._save_raw(
            data=self.start_dark,
            entry=self.start_dark_entry,
            file_path=self.start_dark_file,
        )

        self.end_dark = (
            numpy.random.random((self.dim * self.dim * 2))
            .reshape(2, self.dim, self.dim)
            .astype("f")
        )
        self.end_dark_file = os.path.join(self.tmpdir, "dark.hdf5")
        self.end_dark_entry = "data2"
        self.end_dark_url = self._save_raw(
            data=self.end_dark, entry=self.end_dark_entry, file_path=self.end_dark_file
        )

        # create flats
        self.start_flat = (
            numpy.random.random((self.dim * self.dim * 3))
            .reshape(3, self.dim, self.dim)
            .astype("f")
        )
        self.start_flat_file = os.path.join(self.tmpdir, "start_flat.hdf5")
        self.start_flat_entry = "/root/flat"
        self.start_flat_url = self._save_raw(
            data=self.start_flat,
            entry=self.start_flat_entry,
            file_path=self.start_flat_file,
        )

        self.end_flat = (
            numpy.random.random((self.dim * self.dim))
            .reshape(1, self.dim, self.dim)
            .astype("f")
        )
        # save the end flat in the simple case file to insure all cases are
        # consider
        self.end_flat_file = self._simple_nx.master_file
        self.end_flat_entry = "flat"
        self.end_flat_url = self._save_raw(
            data=self.end_flat, entry=self.end_flat_entry, file_path=self.end_flat_file
        )

    def _save_raw(self, data, entry, file_path) -> DataUrl:
        with HDF5File(file_path, mode="a") as h5s:
            h5s[entry] = data
        return DataUrl(file_path=file_path, data_path=entry, scheme="silx")

    def _create_vds(
        self,
        source_file: str,
        source_data_path: str,
        target_file: str,
        target_data_path: str,
        copy_other_data: bool,
    ):
        """Create virtual dataset and links from source to target

        :param source_file:
        :param source_data_path:
        :param target_file:
        :param target_data_path:
        :param copy_other_data: we want to create two cases: one copying
                                     datasets 'image_key'... and the other
                                     one linking them. Might have a difference
                                     of behavior when overwriting for example
        """
        assert source_file != target_file, "file should be different"
        # link data
        n_frames = 0
        # for now we only consider the original data
        with HDF5File(source_file, mode="r", swmr=get_swmr_mode()) as o_h5s:
            old_path = os.path.join(source_data_path, "instrument", "detector", "data")
            n_frames += o_h5s[old_path].shape[0]
            shape = o_h5s[old_path].shape
            data_type = o_h5s[old_path].dtype

            layout = h5py.VirtualLayout(shape=shape, dtype=data_type)
            assert os.path.exists(source_file)
            with HDF5File(source_file, mode="r", swmr=get_swmr_mode()) as ppp:
                assert source_data_path in ppp
            layout[:] = h5py.VirtualSource(path_or_dataset=o_h5s[old_path])

            det_path = os.path.join(target_data_path, "instrument", "detector")
            with HDF5File(target_file, mode="a") as h5s:
                detector_node = h5s.require_group(det_path)
                detector_node.create_virtual_dataset("data", layout, fillvalue=-5)

        for path in (
            os.path.join("instrument", "detector", "image_key"),
            os.path.join("instrument", "detector", "image_key_control"),
            os.path.join("instrument", "detector", "count_time"),
            os.path.join("sample", "rotation_angle"),
        ):
            old_path = os.path.join(source_data_path, path)
            new_path = os.path.join(target_data_path, path)
            with HDF5File(target_file, mode="a") as h5s:
                if copy_other_data:
                    with HDF5File(source_file, mode="r", swmr=get_swmr_mode()) as o_h5s:
                        if new_path in h5s:
                            del h5s[new_path]
                        h5s[new_path] = h5py_read_dataset(o_h5s[old_path])
                elif source_file == target_file:
                    h5s[new_path] = h5py.SoftLink(old_path)
                else:
                    relpath = os.path.relpath(source_file, os.path.dirname(target_file))
                    h5s[new_path] = h5py.ExternalLink(relpath, old_path)

    def _patch_nxtomo_flags(self, scan):
        """Insure necessary flags are here"""
        with HDF5File(scan.master_file, mode="a") as h5s:
            instrument_path = os.path.join(scan.entry, "instrument")
            instrument_node = h5s.require_group(instrument_path)
            if "NX_class" not in instrument_node.attrs:
                instrument_node.attrs["NX_class"] = "NXinstrument"
            detector_node = instrument_node.require_group("detector")
            if "NX_class" not in detector_node.attrs:
                detector_node.attrs["NX_class"] = "NXdetector"
            if "data" in instrument_node:
                if "interpretation" not in instrument_node.attrs:
                    instrument_node["data"].attrs["interpretation"] = "image"

            sample_path = os.path.join(scan.entry, "sample")
            sample_node = h5s.require_group(sample_path)
            if "NX_class" not in sample_node:
                sample_node.attrs["NX_class"] = "NXsample"

    def tearDown(self) -> None:
        shutil.rmtree(self.tmpdir)


class TestAddDarkAtStart(BaseTestAddDarkAndFlats):
    """
    Make sure adding dark works
    """

    def testAddDarkAsNumpyArray(self) -> None:
        """Insure adding a dark works from a numpy array"""
        for scan in (self._simple_nx,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    darks_start=self.start_dark,
                )

                # test `data` dataset
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape, (self.nproj + 1, self.dim, self.dim)
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self.start_dark[0]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][1], self._raw_data[0]
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)

        # test some exception are raised if we try to add directly a numpy
        # array within a virtual dataset
        for scan in (
            self._nx_with_virtual_dataset,
            self._nx_with_virtual_dataset_with_link,
        ):
            with self.assertRaises(TypeError):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    darks_start=self.start_dark,
                )

    def testAddDarkAsDataUrl(self) -> None:
        """Insure adding a dark works from a DataUrl"""
        for scan in (self._nx_with_virtual_dataset_with_link,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    darks_start=self.start_dark_url,
                )

                # test `data` dataset
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape, (self.nproj + 1, self.dim, self.dim)
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self.start_dark[0]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][1], self._raw_data[0]
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)
                    # test rotation angle and count_time
                    count_time_path = os.path.join(
                        scan.entry, "instrument", "detector", "count_time"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[count_time_path][-1]), 1
                    )
                    self.assertEqual(
                        len(h5s[count_time_path]), self.nproj + self.start_dark.shape[0]
                    )
                    rotation_angle_path = os.path.join(
                        scan.entry, "sample", "rotation_angle"
                    )
                    numpy.testing.assert_array_equal(
                        h5s[rotation_angle_path][0], h5s[rotation_angle_path][1]
                    )
                    self.assertEqual(
                        len(h5s[rotation_angle_path]),
                        self.nproj + self.start_dark.shape[0],
                    )
                    control_data_path = os.path.join(scan.entry, "control", "data")
                    numpy.testing.assert_array_equal(
                        h5s[control_data_path][0], h5s[control_data_path][1]
                    )
                    self.assertEqual(
                        len(h5s[control_data_path]),
                        self.nproj + self.start_dark.shape[0],
                    )


class TestAddFlatAtStart(BaseTestAddDarkAndFlats):
    """
    Make sure adding initial flat works
    """

    def testAddFlatStartAsNumpyArray(self) -> None:
        """Insure adding a dark works from a numpy array"""
        for scan in (self._simple_nx,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_start=self.start_flat,
                )

                # test `data` dataset
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape,
                        (self.nproj + self.start_flat.shape[0], self.dim, self.dim),
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self.start_flat[0]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][2], self.start_flat[2]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][3], self._raw_data[0]
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)

        # test some exception are raised if we try to add directly a numpy
        # array within a virtual dataset
        for scan in (
            self._nx_with_virtual_dataset,
            self._nx_with_virtual_dataset_with_link,
        ):
            with self.assertRaises(TypeError):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_start=self.start_flat,
                )

    def testAddFlatStartAsDataUrl(self) -> None:
        """Insure adding a dark works from a DataUrl"""
        for scan in (self._nx_with_virtual_dataset_with_link,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_start=self.start_flat_url,
                )

                # test `data` dataset
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape,
                        (self.nproj + self.start_flat.shape[0], self.dim, self.dim),
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self.start_flat[0]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][2], self.start_flat[2]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][3], self._raw_data[0]
                    )
                    # test rotation angle and count_time
                    count_time_path = os.path.join(
                        scan.entry, "instrument", "detector", "count_time"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[count_time_path][-1]), 1
                    )
                    self.assertEqual(
                        len(h5s[count_time_path]), self.nproj + self.start_flat.shape[0]
                    )
                    rotation_angle_path = os.path.join(
                        scan.entry, "sample", "rotation_angle"
                    )
                    numpy.testing.assert_array_equal(
                        h5s[rotation_angle_path][0], h5s[rotation_angle_path][1]
                    )
                    self.assertEqual(
                        len(h5s[rotation_angle_path]),
                        self.nproj + self.start_flat.shape[0],
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)


class TestAddFlatAtEnd(BaseTestAddDarkAndFlats):
    """
    Make sure adding final flat works
    """

    def testAddFlatEndAsNumpyArray(self) -> None:
        """Insure adding a dark works from a numpy array"""
        for scan in (self._simple_nx,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_end=self.end_flat,
                )

                # test `data` dataset
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape,
                        (self.nproj + self.end_flat.shape[0], self.dim, self.dim),
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][-1], self.end_flat[-1]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self._raw_data[0]
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)

                    # test the 'image_key' and image_key_control dataset
                    img_key_control_path = os.path.join(
                        scan.entry, "instrument", "detector", "image_key_control"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[img_key_control_path][-2:]), [0, 1]
                    )
                    img_key_path = os.path.join(
                        scan.entry, "instrument", "detector", "image_key"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[img_key_path][-2:]), [0, 1]
                    )
                    # test rotation angle and count_time
                    count_time_path = os.path.join(
                        scan.entry, "instrument", "detector", "count_time"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[count_time_path][-1]), 1
                    )
                    self.assertEqual(
                        len(h5s[count_time_path]), self.nproj + self.end_flat.shape[0]
                    )
                    rotation_angle_path = os.path.join(
                        scan.entry, "sample", "rotation_angle"
                    )
                    numpy.testing.assert_array_equal(
                        h5s[rotation_angle_path][-1], h5s[rotation_angle_path][-2]
                    )
                    self.assertEqual(
                        len(h5s[rotation_angle_path]),
                        self.nproj + self.end_flat.shape[0],
                    )

        # test some exception are raised if we try to add directly a numpy
        # array within a virtual dataset
        for scan in (
            self._nx_with_virtual_dataset,
            self._nx_with_virtual_dataset_with_link,
        ):
            with self.assertRaises(TypeError):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_end=self.end_flat,
                )

    def testAddFlatStartAsDataUrl(self) -> None:
        """Insure adding a dark works from a DataUrl"""
        for scan in (self._nx_with_virtual_dataset_with_link,):
            with self.subTest(scan=scan):
                add_dark_flat_nx_file(
                    file_path=scan.master_file,
                    entry=scan.entry,
                    flats_end=self.end_flat_url,
                )

                # test data is correctly store
                with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
                    # test the 'data' dataset
                    data_path = os.path.join(
                        scan.entry, "instrument", "detector", "data"
                    )
                    self.assertEqual(
                        h5s[data_path].shape,
                        (self.nproj + self.end_flat.shape[0], self.dim, self.dim),
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][-1], self.end_flat[-1]
                    )
                    numpy.testing.assert_array_almost_equal(
                        h5s[data_path][0], self._raw_data[0]
                    )
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)
                    # test the 'image_key' and image_key_control dataset
                    img_key_control_path = os.path.join(
                        scan.entry, "instrument", "detector", "image_key_control"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[img_key_control_path][-2:]), [0, 1]
                    )
                    img_key_path = os.path.join(
                        scan.entry, "instrument", "detector", "image_key"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[img_key_path][-2:]), [0, 1]
                    )
                    # test rotation angle and count_time
                    count_time_path = os.path.join(
                        scan.entry, "instrument", "detector", "count_time"
                    )
                    numpy.testing.assert_array_equal(
                        h5py_read_dataset(h5s[count_time_path][-1]), 1
                    )
                    self.assertEqual(
                        len(h5s[count_time_path]), self.nproj + self.end_flat.shape[0]
                    )
                    rotation_angle_path = os.path.join(
                        scan.entry, "sample", "rotation_angle"
                    )
                    numpy.testing.assert_array_equal(
                        h5s[rotation_angle_path][-1], h5s[rotation_angle_path][-2]
                    )
                    self.assertEqual(
                        len(h5s[rotation_angle_path]),
                        self.nproj + self.end_flat.shape[0],
                    )
                    # insure we can still make a HDF5Scan out of it
                    NXtomoScan(scan=scan.master_file, entry=scan.entry)


class TestAddDarkAndFlatWithFancySelection(BaseTestAddDarkAndFlats):
    """Insure we can do some fancy selection with virtual dataset and data url"""

    def testValid(self):
        """
        Insure virtual dataset are still correctly recreate even if
        using slices
        """
        scan = self._nx_with_virtual_dataset_with_link

        start_flat_url = DataUrl(
            file_path=self.start_flat_file,
            data_path=self.start_flat_entry,
            data_slice=slice(0, 4),
            scheme="silx",
        )
        end_dark_url = DataUrl(
            file_path=self.end_dark_file,
            data_path=self.end_dark_entry,
            data_slice=[
                1,
            ],
            scheme="silx",
        )

        add_dark_flat_nx_file(
            file_path=scan.master_file,
            entry=scan.entry,
            flats_start=start_flat_url,
            darks_end=end_dark_url,
            embed_data=False,
        )

        # test data is correctly store
        with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            # test the 'data' dataset
            data_path = os.path.join(scan.entry, "instrument", "detector", "data")
            self.assertEqual(
                h5s[data_path].shape,
                (self.nproj + 3 + 1, self.dim, self.dim),
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][0:3], self.start_flat[0:3]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-1], self.end_dark[1]
            )
            NXtomoScan(scan=scan.master_file, entry=scan.entry)
            # test the 'image_key' and image_key_control dataset
            img_key_control_path = os.path.join(
                scan.entry, "instrument", "detector", "image_key_control"
            )
            numpy.testing.assert_array_equal(
                h5py_read_dataset(h5s[img_key_control_path][0:3]), [1, 1, 1]
            )
            numpy.testing.assert_array_equal(
                h5py_read_dataset(h5s[img_key_control_path][-2:]), [0, 2]
            )
            # insure we can still make a HDF5Scan out of it
            NXtomoScan(scan=scan.master_file, entry=scan.entry)

    def testInValidSlice(self):
        """Insure an error is raise if user try to
        provide a slice with step !=1. This case is not handled
        """
        scan = self._nx_with_virtual_dataset_with_link

        start_flat_url = DataUrl(
            file_path=self.start_flat_file,
            data_path=self.start_flat_entry,
            data_slice=slice(0, 4, 2),
            scheme="silx",
        )
        end_dark_url = DataUrl(
            file_path=self.end_dark_file,
            data_path=self.end_dark_entry,
            data_slice=[
                1,
            ],
            scheme="silx",
        )
        with self.assertRaises(ValueError):
            add_dark_flat_nx_file(
                file_path=scan.master_file,
                entry=scan.entry,
                flats_start=start_flat_url,
                darks_end=end_dark_url,
                embed_data=False,
            )


class TestCompleteAddFlatAndDark(BaseTestAddDarkAndFlats):
    """
    Complete test on adding dark and flats on a complete case
    """

    def testWithoutExtras(self):
        """Insure a complete case can be handle without defining any extras
        parameters"""
        scan = self._nx_with_virtual_dataset_with_link

        add_dark_flat_nx_file(
            file_path=scan.master_file,
            entry=scan.entry,
            flats_start=self.start_flat_url,
            flats_end=self.end_flat_url,
            darks_start=self.start_dark_url,
            extras={},
        )
        with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            # test the 'data' dataset
            data_path = os.path.join(scan.entry, "instrument", "detector", "data")
            th_shape = (
                self.nproj
                + self.end_flat.shape[0]
                + self.start_flat.shape[0]
                + self.start_dark.shape[0],
                self.dim,
                self.dim,
            )
            self.assertEqual(h5s[data_path].shape, th_shape)

            # test the 'image_key' and image_key_control dataset
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][0], self.start_dark[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][1], self.start_flat[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-1], self.end_flat[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-self.end_flat.shape[0] - 1], self._raw_data[-1]
            )

            # test image_key
            img_key_path = os.path.join(
                scan.entry, "instrument", "detector", "image_key"
            )
            th_img_keys = [0] * self.nproj
            [th_img_keys.insert(0, 1) for _ in range(self.start_flat.shape[0])]
            [th_img_keys.insert(0, 2) for _ in range(self.start_dark.shape[0])]
            [th_img_keys.append(1) for _ in range(self.end_flat.shape[0])]

            numpy.testing.assert_array_equal(h5s[img_key_path], th_img_keys)
            # test rotation_angle dataset
            rotation_angle_dataset = os.path.join(
                scan.entry, "sample", "rotation_angle"
            )
            numpy.testing.assert_array_almost_equal(
                h5s[rotation_angle_dataset][0], h5s[rotation_angle_dataset][1]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[rotation_angle_dataset][0], h5s[rotation_angle_dataset][2]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[rotation_angle_dataset][-2], h5s[rotation_angle_dataset][-1]
            )

    def testWithExtras(self):
        """Insure a complete case can be handle providing some extras
        parameters"""
        scan = self._nx_with_virtual_dataset
        dark_extras = {"rotation_angle": 10.2}
        start_flat_extras = {"rotation_angle": [10, 11, 12]}
        end_flat_extras = {"count_time": 0.3}
        end_dark_extras = {"count_time": [0.1, 0.2], "rotation_angle": [89, 88]}

        extras = {
            "darks_start": dark_extras,
            "darks_end": end_dark_extras,
            "flats_start": start_flat_extras,
            "flats_end": end_flat_extras,
        }
        add_dark_flat_nx_file(
            file_path=scan.master_file,
            entry=scan.entry,
            flats_start=self.start_flat_url,
            flats_end=self.end_flat_url,
            darks_start=self.start_dark_url,
            darks_end=self.end_dark_url,
            extras=extras,
        )
        with HDF5File(scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            # test the 'data' dataset
            data_path = os.path.join(scan.entry, "instrument", "detector", "data")
            th_shape = (
                self.nproj
                + self.end_flat.shape[0]
                + self.start_flat.shape[0]
                + self.start_dark.shape[0]
                + self.end_dark.shape[0],
                self.dim,
                self.dim,
            )
            self.assertEqual(h5s[data_path].shape, th_shape)

            # test the 'image_key' and image_key_control dataset
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][0], self.start_dark[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][1], self.start_flat[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-1], self.end_flat[0]
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-self.end_flat.shape[0] - self.end_dark.shape[0] - 1],
                self._raw_data[-1],
            )
            numpy.testing.assert_array_almost_equal(
                h5s[data_path][-3], self.end_dark[-2]
            )

            # test image_key
            img_key_path = os.path.join(
                scan.entry, "instrument", "detector", "image_key"
            )
            th_img_keys = [0] * self.nproj
            [th_img_keys.insert(0, 1) for _ in range(self.start_flat.shape[0])]
            [th_img_keys.insert(0, 2) for _ in range(self.start_dark.shape[0])]
            [th_img_keys.append(2) for _ in range(self.end_dark.shape[0])]
            [th_img_keys.append(1) for _ in range(self.end_flat.shape[0])]

            numpy.testing.assert_array_equal(h5s[img_key_path], th_img_keys)
            # test rotation_angle dataset
            rotation_angle_dataset = os.path.join(
                scan.entry, "sample", "rotation_angle"
            )
            numpy.testing.assert_array_almost_equal(
                h5s[rotation_angle_dataset][0], 10.2
            )
            numpy.testing.assert_array_almost_equal(h5s[rotation_angle_dataset][-3], 89)
            numpy.testing.assert_array_almost_equal(
                h5py_read_dataset(h5s[rotation_angle_dataset][1:4]),
                numpy.array([10, 11, 12]),
            )
            numpy.testing.assert_array_almost_equal(
                h5s[rotation_angle_dataset][-2], h5s[rotation_angle_dataset][-1]
            )
            count_time_dataset = os.path.join(
                scan.entry, "instrument", "detector", "count_time"
            )
            self.assertEqual(h5s[count_time_dataset][-1], 0.3)
            self.assertEqual(h5s[count_time_dataset][-2], 0.2)
            self.assertEqual(h5s[count_time_dataset][0], 1)


class TestChangeImageKeyControl(unittest.TestCase):
    """
    Test the `change_image_key_control` function
    """

    def setUp(self) -> None:
        self.tmpdir = tempfile.mkdtemp()
        simple_nx_path = os.path.join(self.tmpdir, "simple_case")
        self.dim = 55
        self.nproj = 20
        # this can will have one dark, then 4 flats then 20 projections
        # then 4 flats and 5 alignment projections at the end
        mock = MockNXtomo(
            scan_path=simple_nx_path,
            n_proj=self.nproj,
            n_ini_proj=self.nproj,
            create_ini_dark=True,
            create_ini_flat=True,
            create_final_flat=True,
            dim=self.dim,
            n_refs=4,
        )
        n_alignment = 5
        for i in range(n_alignment):
            mock.add_alignment_radio(i, angle=0)
        self.scan = mock.scan

    def tearDown(self) -> None:
        if os.path.exists(self.tmpdir):
            shutil.rmtree(self.tmpdir)

    def testInputType(self):
        """Check a TypeError is raised if input of `frames_indexes` is
        invalid"""
        with self.assertRaises(TypeError):
            change_image_key_control(
                file_path=self.scan.master_file,
                entry=self.scan.entry,
                frames_indexes=1,
                image_key_control_value=ImageKey.PROJECTION,
            )

    def testUpdateToProjections(self):
        """Insure we can correctly set some frame as `projection`"""
        change_image_key_control(
            file_path=self.scan.master_file,
            entry=self.scan.entry,
            frames_indexes=slice(0, 3, 2),
            image_key_control_value=ImageKey.PROJECTION,
        )
        with HDF5File(self.scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            image_keys_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key")
            )
            image_keys = h5s[image_keys_path]
            image_keys_control_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key_control")
            )
            self.assertEqual(image_keys[0], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys[1], ImageKey.FLAT_FIELD.value)
            self.assertEqual(image_keys[2], ImageKey.PROJECTION.value)

            image_keys_control = h5s[image_keys_control_path]
            self.assertEqual(image_keys_control[0], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys_control[1], ImageKey.FLAT_FIELD.value)
            self.assertEqual(image_keys_control[2], ImageKey.PROJECTION.value)

    def testUpdateToDark(self):
        """Insure we can correctly set some frame as `dark`"""
        change_image_key_control(
            file_path=self.scan.master_file,
            entry=self.scan.entry,
            frames_indexes=[1],
            image_key_control_value=ImageKey.DARK_FIELD,
        )
        with HDF5File(self.scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            image_keys_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key")
            )
            image_keys = h5s[image_keys_path]
            image_keys_control_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key_control")
            )
            self.assertEqual(image_keys[0], ImageKey.DARK_FIELD.value)
            self.assertEqual(image_keys[1], ImageKey.DARK_FIELD.value)
            self.assertEqual(image_keys[2], ImageKey.FLAT_FIELD.value)

            image_keys_control = h5s[image_keys_control_path]
            self.assertEqual(image_keys_control[0], ImageKey.DARK_FIELD.value)
            self.assertEqual(image_keys_control[1], ImageKey.DARK_FIELD.value)
            self.assertEqual(image_keys_control[2], ImageKey.FLAT_FIELD.value)

    def testUpdateToFlat(self):
        """Insure we can correctly set some frame as `flatfield`"""
        change_image_key_control(
            file_path=self.scan.master_file,
            entry=self.scan.entry,
            frames_indexes=[0],
            image_key_control_value=ImageKey.FLAT_FIELD,
        )
        with HDF5File(self.scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            image_keys_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key")
            )
            image_keys = h5s[image_keys_path]
            image_keys_control_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key_control")
            )
            self.assertEqual(image_keys[0], ImageKey.FLAT_FIELD.value)
            self.assertEqual(image_keys[1], ImageKey.FLAT_FIELD.value)

            image_keys_control = h5s[image_keys_control_path]
            self.assertEqual(image_keys_control[0], ImageKey.FLAT_FIELD.value)
            self.assertEqual(image_keys_control[1], ImageKey.FLAT_FIELD.value)

    def testUpdateToAlignment(self):
        """Insure we can correctly set some frame as `alignment`"""
        change_image_key_control(
            file_path=self.scan.master_file,
            entry=self.scan.entry,
            frames_indexes=slice(10, 20, None),
            image_key_control_value=ImageKey.ALIGNMENT,
        )
        with HDF5File(self.scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            image_keys_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key")
            )
            image_keys = h5s[image_keys_path]
            image_keys_control_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key_control")
            )
            self.assertEqual(image_keys[10], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys[11], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys[20], ImageKey.PROJECTION.value)

            image_keys_control = h5s[image_keys_control_path]
            self.assertEqual(image_keys_control[9], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys_control[10], ImageKey.ALIGNMENT.value)
            self.assertEqual(image_keys_control[11], ImageKey.ALIGNMENT.value)
            self.assertEqual(image_keys_control[21], ImageKey.PROJECTION.value)

    def testUpdateToInvalid(self):
        """Insure we can correctly set some frame as `invalid`"""
        change_image_key_control(
            file_path=self.scan.master_file,
            entry=self.scan.entry,
            frames_indexes=slice(0, 16, 5),
            image_key_control_value=ImageKey.INVALID,
        )
        with HDF5File(self.scan.master_file, mode="r", swmr=get_swmr_mode()) as h5s:
            image_keys_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key")
            )
            image_keys = h5s[image_keys_path]
            image_keys_control_path = "/".join(
                (self.scan.entry, "instrument", "detector", "image_key_control")
            )
            self.assertEqual(image_keys[9], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys[10], ImageKey.INVALID.value)
            self.assertEqual(image_keys[11], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys[15], ImageKey.INVALID.value)

            image_keys_control = h5s[image_keys_control_path]
            self.assertEqual(image_keys_control[9], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys_control[10], ImageKey.INVALID.value)
            self.assertEqual(image_keys_control[11], ImageKey.PROJECTION.value)
            self.assertEqual(image_keys_control[15], ImageKey.INVALID.value)


class TestAddDarkAndFlatFromADifferentFolderWithVDS(unittest.TestCase):
    """
    Test adding dark and flat to vds pointing to vds pointing on files at
    at different level of the file system.
                               root_folder
            _______________________|_____________
           |                                    |
       folder_1                              folder_2
           |                                    |
      file created      _________________________|___________________
                       |               |                            |
                  subfolder_21     subfolder_22                 file with
                       |               |                    original start dark
               File containing     ____|___________
              VDS pointing to     |               |
              original flat    File with     subsubfolder_221
              and dark         original              |
                               start flat   File with original
                                               end flat
    """

    def setUp(self) -> None:
        unittest.TestCase.setUp(self)
        self.dim = 20
        # create folder
        self.root_folder = tempfile.mkdtemp()
        self.mv_folder = tempfile.mkdtemp()
        self.folder_1 = os.path.join(self.root_folder, "folder_1")
        self.folder_2 = os.path.join(self.root_folder, "folder_2")
        self.subfolder_21 = os.path.join(self.root_folder, "subfolder_21")
        self.subfolder_22 = os.path.join(self.root_folder, "subfolder_22")
        self.subsubfolder_221 = os.path.join(self.root_folder, "subsubfolder_221")
        for folder in (
            self.folder_1,
            self.folder_2,
            self.subfolder_21,
            self.subfolder_22,
            self.subsubfolder_221,
        ):
            os.makedirs(folder)

        # create original dark
        self.start_dark = (
            numpy.random.random((5 * self.dim * self.dim))
            .reshape(5, self.dim, self.dim)
            .astype("f")
        )
        self.start_dark_file = os.path.join(self.folder_2, "original_start_dark.hdf5")
        with HDF5File(self.start_dark_file, mode="w") as h5s:
            h5s["dark"] = self.start_dark
        self.start_dark_url = DataUrl(
            file_path=self.start_dark_file, data_path="dark", scheme="silx"
        )

        # create original flats
        self.start_flat = (
            numpy.random.random((5 * self.dim * self.dim))
            .reshape(5, self.dim, self.dim)
            .astype("f")
        )
        self.start_flat_file = os.path.join(self.subfolder_22, "original_start_flat.h5")
        with HDF5File(self.start_flat_file, mode="w") as h5s:
            h5s["flat1"] = self.start_flat
        self.start_flat_url = DataUrl(
            file_path=self.start_flat_file, data_path="flat1", scheme="silx"
        )

        self.end_flat = (
            numpy.random.random((5 * self.dim * self.dim))
            .reshape(5, self.dim, self.dim)
            .astype("f")
        )
        self.end_flat_file = os.path.join(self.subsubfolder_221, "original_end_flat.h5")
        with HDF5File(self.end_flat_file, mode="w") as h5s:
            h5s["flat2"] = self.end_flat
        self.end_flat_url = DataUrl(
            file_path=self.end_flat_file, data_path="flat2", scheme="silx"
        )

    def tearDown(self) -> None:
        for folder in (self.root_folder, self.mv_folder):
            if os.path.exists(folder):
                shutil.rmtree(folder)
        unittest.TestCase.tearDown(self)

    def test(self):
        # create a scan contained in subfolder21
        scan_subfolder_21 = self._simple_nx = MockNXtomo(
            scan_path=self.subfolder_21,
            n_proj=10,
            n_ini_proj=2,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=self.dim,
        ).scan
        # 1. add first level of indirection for dark and flat and check the VDS
        add_dark_flat_nx_file(
            file_path=scan_subfolder_21.master_file,
            entry=scan_subfolder_21.entry,
            flats_start=self.start_flat_url,
            flats_end=self.end_flat_url,
            darks_start=self.start_dark_url,
            extras={
                "darks_start": {"count_time": 0.3, "rotation_angle": 0.0},
                "flats_start": {"count_time": 0.3, "rotation_angle": 0.0},
                "flats_end": {"count_time": 0.3, "rotation_angle": 0.0},
            },
        )
        self.check_vds_from_file(
            file_=scan_subfolder_21.master_file, nx_entry=scan_subfolder_21.entry
        )

        # create a scan contained in subfolder221
        scan_subfolder_221 = self._simple_nx = MockNXtomo(
            scan_path=self.subsubfolder_221,
            n_proj=10,
            n_ini_proj=2,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=self.dim,
        ).scan
        # 2. create a second level of indirection for dark and flat and check
        # the VDS
        det_data_path = "/".join(
            (scan_subfolder_21.entry, "instrument", "detector", "data")
        )
        new_dark_url = DataUrl(
            file_path=scan_subfolder_21.master_file,
            data_path=det_data_path,
            data_slice=slice(0, 5),
            scheme="silx",
        )
        new_start_flat_url = DataUrl(
            file_path=scan_subfolder_21.master_file,
            data_path=det_data_path,
            data_slice=slice(5, 10),
            scheme="silx",
        )
        new_end_flat_url = DataUrl(
            file_path=scan_subfolder_21.master_file,
            data_path=det_data_path,
            data_slice=slice(12, 17),
            scheme="silx",
        )

        add_dark_flat_nx_file(
            file_path=scan_subfolder_221.master_file,
            entry=scan_subfolder_221.entry,
            flats_start=new_start_flat_url,
            flats_end=new_end_flat_url,
            darks_start=new_dark_url,
            extras={
                "darks_start": {"count_time": 0.3, "rotation_angle": 0.0},
                "flats_start": {"count_time": 0.3, "rotation_angle": 0.0},
                "flats_end": {"count_time": 0.3, "rotation_angle": 0.0},
            },
        )
        self.check_vds_from_file(
            file_=scan_subfolder_221.master_file, nx_entry=scan_subfolder_221.entry
        )

        # 3. move root folder and remove it to insure links are style valid
        shutil.move(src=self.folder_1, dst=self.mv_folder)
        shutil.move(src=self.folder_2, dst=self.mv_folder)
        new_scan_subfolder_221_m = scan_subfolder_221.master_file.replace(
            self.mv_folder, self.root_folder
        )

        self.check_vds_from_file(
            file_=new_scan_subfolder_221_m, nx_entry=scan_subfolder_221.entry
        )

    def check_vds_from_file(self, file_, nx_entry):
        with HDF5File(file_, mode="r", swmr=get_swmr_mode()) as h5s:
            entry_node = h5s[nx_entry]
            det_group = entry_node["instrument/detector"]
            det_data = det_group["data"]
            det_image_key = det_group["image_key"]

            # test dark
            numpy.testing.assert_array_equal(
                det_image_key[:5], [ImageKey.DARK_FIELD.value] * 5
            )
            numpy.testing.assert_array_equal(det_data[:5], self.start_dark)
            # test start flat
            numpy.testing.assert_array_equal(
                det_image_key[5:10], [ImageKey.FLAT_FIELD.value] * 5
            )
            numpy.testing.assert_array_equal(det_data[5:10], self.start_flat)
            # test end flat
            numpy.testing.assert_array_equal(
                det_image_key[-5:], [ImageKey.FLAT_FIELD.value] * 5
            )
            numpy.testing.assert_array_equal(det_data[-5:], self.end_flat)



"""Util to load dataset from gitlab.esrf.fr:tomo/tomo_test_data"""

import os
import pytest
import shutil
from tomoscan.tests.datasets import GitlabProject as _GitlabProject

_GitlabBlissTomoHolotomoDataset = _GitlabProject(
    branch_name="holotomo2",
    host="https://gitlab.esrf.fr",
    cache_dir=os.path.join(
        os.path.dirname(__file__),
        "__archive__",
    ),
    token=None,
    project_id=4166,  # https://gitlab.esrf.fr/tomo/tomo_test_data
)


@pytest.fixture
def holotomo2_simu_dataset(tmp_path):
    """
    Return a  bliss-tomo simulated dataset with latest holo-tomo sequence.
    """
    src_folder = _GitlabBlissTomoHolotomoDataset.get_dataset(
        os.path.join("simu", "bliss-2.3.0.dev0-tomo-2.9.0", "holotomo2")
    )
    dst_folder = os.path.join(tmp_path, "holotomo2")
    shutil.copytree(src_folder, dst_folder)
    return dst_folder



# coding: utf-8

"""
Application to create a default configuration file to be used by blissfluo2nx application.

.. program-output:: nxtomomill blissfluo-config --help

"""

import argparse
import logging

from nxtomomill.models.blissfluo2nx import BlissFluo2nxModel

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(description="Create a default configuration file")
    parser.add_argument("output_file", help="output .cfg file")
    parser.add_argument(
        "--level",
        "--option-level",
        help="Level of options to embed in the configuration file. Can be 'required' or 'advanced'.",
        default="required",
    )

    options = parser.parse_args(argv[1:])

    configuration = BlissFluo2nxModel()
    configuration.to_cfg_file(file_path=options.output_file)



import numpy


def deduce_machine_current(timestamps: tuple, known_machine_current: dict) -> tuple:
    """
    :param known_machine_current: keys are timestamp. Value is machine current
    :param timestamps: timestamp for which we want to get machine current
    """
    if not isinstance(known_machine_current, dict):
        raise TypeError("known_machine_current is expected to be a dict")
    for elmt in timestamps:
        if not isinstance(elmt, numpy.datetime64):
            raise TypeError(
                f"Elements of timestamps are expected to be {numpy.datetime64} and not {type(elmt)}"
            )
    if len(known_machine_current) == 0:
        raise ValueError("known_machine_current should contain at least one element")
    for key, value in known_machine_current.items():
        if not isinstance(key, numpy.datetime64):
            raise TypeError(
                f"known_machine_current keys are expected to be instances of {numpy.datetime64} and not {type(key)}"
            )
        if not isinstance(value, (float, numpy.number)):
            raise TypeError(
                "known_machine_current values are expected to be instances of float"
            )

    # 1. Order **known** machine current by time stamps (key)
    known_machine_current = dict(sorted(known_machine_current.items()))
    known_timestamps = numpy.fromiter(
        known_machine_current.keys(),
        dtype="datetime64[ns]",
        count=len(known_machine_current),
    )
    known_machine_current_values = numpy.fromiter(
        known_machine_current.values(),
        dtype="float64",
        count=len(known_machine_current),
    )

    # 2. Sort the supplied timestamps
    timestamps = numpy.array(timestamps, dtype="datetime64[ns]")
    timestamp_input_ordering = numpy.argsort(timestamps)
    timestamps_sorted = numpy.take_along_axis(
        timestamps, indices=timestamp_input_ordering, axis=0
    )

    # 3. Convert to float for numpy.interp
    known_timestamps_float = known_timestamps.astype("float64")
    timestamps_float = timestamps_sorted.astype("float64")

    # 4. Interpolate the values
    interpolated_values = numpy.interp(
        timestamps_float, known_timestamps_float, known_machine_current_values
    )

    # 5. Reorder interpolated values to match the order of original timestamps
    ordered_interpolated_values = numpy.zeros_like(interpolated_values)
    for i, o_pos in enumerate(timestamp_input_ordering):
        ordered_interpolated_values[o_pos] = interpolated_values[i]
    return tuple(ordered_interpolated_values)



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator, ConfigDict, field_serializer
from nxtomo.nxobject.nxsource import ProbeType, SourceType
from nxtomomill.models.utils import filter_str_def


class SourceSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    source_name: str | None = Field(
        default="ESRF", description="Name of the instrument"
    )

    source_type: SourceType | None = Field(
        default=SourceType.SYNCHROTRON_X_RAY_SOURCE, description="Source type"
    )

    source_probe: ProbeType | None = Field(
        default=ProbeType.X_RAY, description="Probe type"
    )

    @field_validator(
        "source_type",
        mode="plain",
    )
    @classmethod
    def cast_to_source_type(cls, value: str | SourceType | None) -> SourceType:
        if value in (None, "", "None"):
            return None
        elif isinstance(value, str):
            value = filter_str_def(value)
        return SourceType(value)

    @field_serializer(
        "source_type",
    )
    @classmethod
    def serialize_source_type(cls, source_type: SourceType | None) -> str:
        if source_type is None:
            return ""
        else:
            return source_type.value

    @field_validator(
        "source_probe",
        mode="plain",
    )
    @classmethod
    def cast_to_source_probe(cls, value: str | ProbeType | None) -> ProbeType:
        if value in (None, ""):
            return None
        elif isinstance(value, str):
            value = filter_str_def(value)
        return ProbeType(value)

    @field_serializer(
        "source_probe",
    )
    @classmethod
    def serialize_source_probe(cls, source_probe: ProbeType | None) -> str:
        if source_probe is None:
            return ""
        else:
            return source_probe.value

    @field_validator(
        "source_name",
        mode="plain",
    )
    @classmethod
    def cast_to_str(cls, value: str | None) -> tuple | None:
        return filter_str_def(value)



# coding: utf-8

import os
import tempfile

from tomoscan.esrf.mock import MockNXtomo

from nxtomomill.app.patch_nx import main


def test_patch_nx_application():
    """test nxtomomill patch-nx input_file entry --invalid-frames XX:YY"""

    with tempfile.TemporaryDirectory() as folder:
        nx_path = os.path.join(folder, "nexus_file.nx")
        dim = 55
        nproj = 20
        scan = MockNXtomo(
            scan_path=nx_path,
            n_proj=nproj,
            n_ini_proj=nproj,
            create_ini_dark=False,
            create_ini_flat=False,
            create_final_flat=False,
            dim=dim,
        ).scan
        main(
            [
                "patch-nx",
                scan.master_file,
                "entry",
                "--invalid-frames",
                "0:12",
            ]
        )
        scan.clear_cache()
        assert len(scan.projections) == 8, "Scan is expected to have 8 projections now"



from __future__ import annotations

import h5py


def has_valid_detector(node, detectors_names):
    """
    :return True if the node looks like a valid nx detector
    """
    for key in node.keys():
        if (
            "NX_class" in node[key].attrs
            and node[key].attrs["NX_class"] == "NXdetector"
        ):
            if detectors_names is None or key in detectors_names:
                return True
    return False


def get_nx_detectors(node: h5py.Group) -> tuple:
    """

    :param node: node to inspect
    :return: tuple of NXdetector (h5py.Group) contained in `node`
             (expected to be the `instrument` group)
    """
    if not isinstance(node, h5py.Group):
        raise TypeError("node should be an instance of h5py.Group")
    nx_detectors = []
    for _, subnode in node.items():
        if isinstance(subnode, h5py.Group) and "NX_class" in subnode.attrs:
            if subnode.attrs["NX_class"] == "NXdetector":
                if "data" in subnode and hasattr(subnode["data"], "ndim"):
                    if subnode["data"].ndim == 3:
                        nx_detectors.append(subnode)
    nx_detectors = sorted(nx_detectors, key=lambda det: det.name)
    return tuple(nx_detectors)


def guess_nx_detector(node: h5py.Group) -> tuple:
    """
    Try to guess what can be an nx_detector without using the "NXdetector"
    NX_class attribute. Expect to find a 3D dataset named 'data' under
    a subnode
    """
    if not isinstance(node, h5py.Group):
        raise TypeError("node should be an instance of h5py.Group")
    nx_detectors = []
    for _, subnode in node.items():
        if isinstance(subnode, h5py.Group) and "data" in subnode:
            if isinstance(subnode["data"], h5py.Dataset) and subnode["data"].ndim == 3:
                nx_detectors.append(subnode)

    nx_detectors = sorted(nx_detectors, key=lambda det: det.name)
    return tuple(nx_detectors)



from __future__ import annotations

import os
import pint
import numpy

from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey

from nxtomomill.tests.datasets import GitlabDataset
from nxtomomill import converter
from nxtomomill.io.config import TomoHDF5Config

_ureg = pint.get_application_registry()


def test_back_and_forth_conversion(tmp_path):
    """test nxtomomill fluo2nx CLI with default parameters (handle all detectors found)"""
    scan_dir = GitlabDataset.get_dataset("h5_datasets")
    assert scan_dir is not None
    input_file = os.path.join(scan_dir, "back_and_forth/bliss_tomo_2025_06_17.h5")

    output_file = os.path.join(tmp_path, "nexus_file.nx")

    assert not os.path.exists(output_file), "output_file exists already"
    configuration = TomoHDF5Config()

    configuration.output_file = output_file
    configuration.input_file = input_file
    configuration.no_input = True
    configuration.raises_error = False
    configuration.no_master_file = False

    # Step 1: make sure conversion is not done if we mess with tiles (and make sure this is back-and-forth which is recognized and not another like multi-tomo)
    tmp_back_and_forth_titles = configuration.back_and_forth_init_titles
    configuration.back_and_forth_init_titles = tuple()

    result = converter.from_h5_to_nx(configuration=configuration)
    assert not os.path.exists(output_file)

    # Step 2: test the real conversion using default inputs
    configuration.back_and_forth_init_titles = tmp_back_and_forth_titles
    result = converter.from_h5_to_nx(configuration=configuration)
    assert os.path.exists(output_file)
    # check we have our 3 NXtomo are converted with each 101 projections, a flat and a dark at the beginning and a flat at the end.
    assert len(result) == 3
    nx_tomos = [NXtomo().load(file_path, entry) for (file_path, entry) in result]
    n_proj_per_nx_tomo = 101
    for i_nx_tomo, nx_tomo in enumerate(nx_tomos):
        # check instrument
        assert len(nx_tomo.instrument.detector.image_key) == n_proj_per_nx_tomo + 40 * 2
        # image key
        numpy.testing.assert_array_equal(
            nx_tomo.instrument.detector.image_key,
            numpy.concatenate(
                (
                    [ImageKey.DARK_FIELD] * 20,
                    [ImageKey.FLAT_FIELD] * 20,
                    [ImageKey.PROJECTION] * n_proj_per_nx_tomo,
                    [ImageKey.FLAT_FIELD] * 20,
                    [ImageKey.DARK_FIELD] * 20,
                )
            ),
        )
        # sequence number
        proj_seq_start = n_proj_per_nx_tomo * i_nx_tomo + 40
        second_flat_start_index = 40 + 3 * n_proj_per_nx_tomo
        numpy.testing.assert_array_equal(
            nx_tomo.instrument.detector.sequence_number,
            numpy.concatenate(
                (
                    numpy.arange(0, 20, dtype=numpy.uint32),  # dark
                    numpy.arange(20, 40, dtype=numpy.uint32),  # flat
                    numpy.arange(
                        proj_seq_start,
                        proj_seq_start + n_proj_per_nx_tomo,
                        dtype=numpy.uint32,
                    ),  # projections
                    numpy.arange(
                        second_flat_start_index,
                        second_flat_start_index + 20,
                        dtype=numpy.uint32,
                    ),  # flat
                    numpy.arange(
                        second_flat_start_index + 20,
                        second_flat_start_index + 40,
                        dtype=numpy.uint32,
                    ),  # dark
                )
            ),
        )
        # detector pixel size
        numpy.testing.assert_almost_equal(
            nx_tomo.instrument.detector.x_pixel_size.magnitude,
            (2.55135 * _ureg.micrometer).magnitude,
            decimal=5,
        )
        numpy.testing.assert_almost_equal(
            nx_tomo.instrument.detector.y_pixel_size.magnitude,
            (2.55135 * _ureg.micrometer).magnitude,
            decimal=5,
        )

        assert nx_tomo.instrument.detector.distance == 3500 * _ureg.millimeter
        # check energy
        assert nx_tomo.energy == 10.0 * _ureg.keV
        # check sample
        assert nx_tomo.sample.x_pixel_size == 2.4 * _ureg.micrometer
        assert nx_tomo.sample.y_pixel_size == 2.4 * _ureg.micrometer
        # check source
        assert nx_tomo.instrument.source.distance == -55500 * _ureg.millimeter



"""data exchange (dx) configuration module"""

from __future__ import annotations

from nxtomo.nxobject.nxdetector import FieldOfView
from .base.general_section import GeneralSection
from pydantic import Field, ConfigDict, field_serializer, field_validator

__all__ = [
    "DX2nxModel",
]


class DX2nxModel(GeneralSection):
    """Configuration file to run a conversion from data-exchange file to NXtomo format"""

    model_config: ConfigDict = ConfigDict(validate_assignment=True)

    input_file: str = Field(
        description="Path to input file at data-file-exchange format to be converted to NXtomo format"
    )
    copy_data: bool = Field(
        default=True,
        description="If True frames will be duplicated. Otherwise we will create (relative) link to the input file.",
    )
    input_entry: str = Field(
        default="/",
        description="Path to the HDF5 group to convert. For now it looks each file can only contain one dataset. Just to ensure any future compatibility if it evolve with time.",
    )
    output_entry: str = Field(
        "entry0000", description="Path to store the NxTomo created."
    )
    scan_range: tuple[float, float] = Field(
        default=(0.0, 180.0),
        description="Tuple of two elements with the minimum scan range. Projections are expected to be taken with equal angular spaces.",
    )
    pixel_size: tuple[float | None, float | None] = Field(
        default=(None, None),
        description="Pixel size can be provided - in meter - as (x_pizel_size, y_pixel_size)",
    )
    field_of_view: FieldOfView | None = Field(
        default=None, description="Detector field of view"
    )
    sample_detector_distance: float | None = Field(
        default=1.0, description="sample / detector distance in meter"
    )
    energy: float | None = Field(default=None, description="Energy in keV")

    @field_validator(
        "field_of_view",
    )
    @classmethod
    def cast_to_field_of_view(cls, value: str | FieldOfView) -> FieldOfView | None:
        if value in (None, ""):
            return None
        return FieldOfView(value)

    @field_serializer(
        "field_of_view",
        when_used="always",
    )
    @classmethod
    def serialize_field_of_view(cls, field_of_view: FieldOfView | None) -> str:
        if field_of_view is None:
            return None
        else:
            return field_of_view.value



from __future__ import annotations

from pydantic import BaseModel, Field, field_validator, ConfigDict


class MultiTomoSection(BaseModel):

    model_config: ConfigDict = ConfigDict(
        validate_assignment=True, validate_by_name=True
    )

    start_angle_offset_in_degree: float | None = Field(
        default=None, description="Start angle offset in degree."
    )
    n_nxtomo: int = Field(default=-1, description="Number of NXtomo to create.")
    angle_interval_in_degree: int = Field(
        default=360, description="Angle interval to create."
    )
    shift_angles: bool = Field(
        default=False, description="Shift all angle NXtomo angle."
    )

    @field_validator(
        "start_angle_offset_in_degree",
        mode="plain",
    )
    @classmethod
    def cast_multi_tomo_section_to_float(cls, value: str) -> float | None:
        if value in (None, "None", ""):
            return None
        return float(value)

    # field aliases

    @property
    def multitomo_start_angle_offset(self):
        return self.start_angle_offset_in_degree

    @multitomo_start_angle_offset.setter
    def multitomo_start_angle_offset(self, value):
        self.start_angle_offset_in_degree = value

    @property
    def multitomo_scan_range(self):
        return self.angle_interval_in_degree

    @multitomo_scan_range.setter
    def multitomo_scan_range(self, value):
        self.angle_interval_in_degree = value

    @property
    def multitomo_shift_angles(self):
        return self.shift_angles

    @multitomo_shift_angles.setter
    def multitomo_shift_angles(self, value):
        self.shift_angles = value

    @property
    def multitomo_n_nxtomo(self):
        return self.n_nxtomo

    @multitomo_n_nxtomo.setter
    def multitomo_n_nxtomo(self, value):
        self.n_nxtomo = value



# coding: utf-8

import os
import pint
from tempfile import TemporaryDirectory

import pytest
import logging

# from pydantic import ValidateError
import importlib.resources as importlib_resources
from configparser import ConfigParser

from nxtomo.nxobject.nxdetector import FieldOfView
from nxtomo.nxobject.nxsource import ProbeType, SourceType

from nxtomomill import settings
from nxtomomill.models.edf2nx import EDF2nxModel, generate_default_edf_config
from nxtomomill.tests.resources import edf2nx_config_files as _edf2nx_config_files
from nxtomomill.utils import FileExtension
from nxtomomill.models.utils import PathType

_ureg = pint.get_application_registry()


def test_moving_to_pydantic_default_config_file():
    """
    test that the default nxtomomill config generated before moving to pydantic is properly interpreted.
    This ignores any comments.
    """
    resources = importlib_resources.files(_edf2nx_config_files.__name__)
    ref_file = os.path.join(resources, "default_file_before_moving_to_pydantic.cfg")
    assert os.path.exists(ref_file)
    txt_parser = ConfigParser(allow_no_value=True)
    txt_parser.read(ref_file)

    old_config = EDF2nxModel(
        general_section=txt_parser["GENERAL_SECTION"],
        edf_keys_section=txt_parser["EDF_KEYS_SECTION"],
        dark_and_flat_section=txt_parser["DARK_AND_FLAT_SECTION"],
        sample_section=txt_parser["SAMPLE_SECTION"],
        unit_section=txt_parser["UNIT_SECTION"],
        source_section=txt_parser["SOURCE_SECTION"],
        detector_section=txt_parser["DETECTOR_SECTION"],
    )
    old_dict = {key.upper(): value for key, value in old_config.model_dump().items()}

    new_dict = generate_default_edf_config()
    assert new_dict == old_dict


def test_TomoEDFConfig_default_config():
    """insure default configuration generation works"""
    with TemporaryDirectory() as folder:
        file_path = os.path.join(folder, "config.cfg")
        assert not os.path.exists(file_path)
        config = generate_default_edf_config()
        assert isinstance(config, dict)
        EDF2nxModel().to_cfg_file(file_path=file_path)
        assert os.path.exists(file_path)


def test_moving_to_pydantic_modified_config_file():
    """
    test that an old nxtomomill edf-config file generated before moving to pydantic is properly interpreted.
    This ignores any comments.
    """
    resources = importlib_resources.files(_edf2nx_config_files.__name__)
    ref_file = os.path.join(resources, "fully_modified_file_before_moving.cfg")
    assert os.path.exists(ref_file)

    config: EDF2nxModel = EDF2nxModel.from_cfg_file(ref_file)
    assert config.input_folder == "my_folder"
    assert config.output_file == "my_file.nx"
    assert config.overwrite is True
    assert config.delete_edf_source_files is True
    assert config.output_checks == ("check_1",)
    assert config.file_extension == FileExtension.NXS
    assert config.dataset_basename == "dataset_basename"
    assert config.dataset_info_file == "dataset_info_file.info"
    assert config.log_level == logging.INFO
    assert config.title == "my_title"
    assert config.patterns_to_ignores == ("_slice_", "_pattern_")
    assert config.duplicate_data is False
    assert config.external_link_type is PathType.ABSOLUTE

    assert config.motor_position_keys == ("motor_pos_key",)
    assert config.motor_mne_keys == ("motor_mne_key",)
    assert config.rotation_angle_keys == ("srotatation",)
    assert config.x_translation_keys == ("stx",)
    assert config.y_translation_keys == ("sty",)
    assert config.z_translation_keys == ("stz",)

    assert config.dark_names_prefix == ("prefix_1", "prefix_2")
    assert config.flat_names_prefix == ("prefix_3", "prefix_4")

    assert config.pixel_size_unit == _ureg.meter
    assert config.sample_detector_distance_unit == _ureg.meter
    assert config.energy_unit == _ureg.GeV
    assert config.x_translation_unit == _ureg.meter
    assert config.y_translation_unit == _ureg.meter
    assert config.z_translation_unit == _ureg.meter
    assert config.machine_current_unit == _ureg.kA

    assert config.sample_name == "my_sample"
    assert config.force_angle_calculation is False
    assert config.angle_calculation_endpoint is True
    assert config.angle_calculation_rev_neg_scan_range is False

    assert config.instrument_name == "my_instrument"
    assert config.source_name == "my_source"
    assert config.source_type == SourceType.SPALLATION_NEUTRON
    assert config.source_probe == ProbeType.NEUTRON

    assert config.field_of_view == FieldOfView.HALF


def test_TomoEDFConfig_setters():
    """test the different setters and getter of the EDFTomoConfig"""
    config = EDF2nxModel()

    # try to test a new attribut (insure class is frozeen)
    with pytest.raises(ValueError):
        config.new_attrs = "toto"

    # test general section setters
    config.input_folder = "my_folder"
    config.output_file = "my_nx.nx"

    config.file_extension = ".nx"

    with pytest.raises(ValueError):
        config.dataset_basename = 12.0
    config.dataset_basename = None
    config.dataset_basename = "test_"

    with pytest.raises(ValueError):
        config.dataset_info_file = 12.3
    config.dataset_info_file = None
    config.dataset_info_file = "my_info_file.info"

    config.overwrite = True

    with pytest.raises(ValueError):
        config.title = 12.0
    config.title = None
    config.title = "my title"

    with pytest.raises(ValueError):
        config.patterns_to_ignores = 12.0
    config.patterns_to_ignores = ("toto",)

    # test header keys + dark and flat setters
    attributes_value = {
        "motor_position_keys": settings.Tomo.EDF.MOTOR_POS,
        "motor_mne_keys": settings.Tomo.EDF.MOTOR_MNE,
        "rotation_angle_keys": settings.Tomo.EDF.ROT_ANGLE,
        "dark_names": settings.Tomo.EDF.DARK_NAMES,
        "flat_names": settings.Tomo.EDF.REFS_NAMES,
    }
    for attr, key in attributes_value.items():
        setattr(config, attr, key)
        with pytest.raises((TypeError, ValueError)):
            setattr(config, attr, "toto")
        with pytest.raises((TypeError, ValueError)):
            setattr(config, attr, 12.20)
        with pytest.raises((TypeError, ValueError)):
            setattr(config, attr, (12.20,))

    # test units setters
    attributes_value = {
        "pixel_size_unit": _ureg.micrometer,
        "distance_unit": _ureg.meter,
        "energy_unit": _ureg.keV,
        "x_translation_unit": _ureg.meter,
        "y_translation_unit": _ureg.meter,
        "z_translation_unit": _ureg.meter,
    }
    for attr, key in attributes_value.items():
        # test providing an instance of a unit
        setattr(config, attr, key)
        # test providing a sting if of a unit
        setattr(config, attr, str(key))
        with pytest.raises(TypeError):
            setattr(config, attr, None)
        with pytest.raises(TypeError):
            setattr(config, attr, 12.0)

    # test sample setters
    config.sample_name = None
    with pytest.raises(ValueError):
        config.sample_name = (12,)
    with pytest.raises(ValueError):
        config.sample_name = 12.0
    config.sample_name = "my sample"

    with pytest.raises(ValueError):
        config.force_angle_calculation = "toto"
    config.force_angle_calculation = True

    with pytest.raises(ValueError):
        config.angle_calculation_endpoint = "toto"
    config.angle_calculation_endpoint = True

    with pytest.raises(ValueError):
        config.angle_calculation_rev_neg_scan_range = "toto"
    config.angle_calculation_rev_neg_scan_range = False

    # test source setters
    config.instrument_name = None
    config.instrument_name = "BMXX"
    with pytest.raises(ValueError):
        config.instrument_name = 12.3

    config.source_name = None
    config.source_name = "ESRF"
    with pytest.raises(ValueError):
        config.source_name = 12.2
    config.source_name = "XFEL"

    config.source_type = None
    config.source_type = SourceType.FIXED_TUBE_X_RAY.value
    config.source_type = SourceType.FIXED_TUBE_X_RAY
    with pytest.raises(ValueError):
        config.source_type = "trsts"
    with pytest.raises(ValueError):
        config.source_type = 123

    config.source_probe = None
    config.source_probe = ProbeType.ELECTRON
    config.source_probe = "positron"
    with pytest.raises(ValueError):
        config.source_probe = 123

    # test detector setters
    config.field_of_view = None
    with pytest.raises(ValueError):
        config.field_of_view = 12
    with pytest.raises(ValueError):
        config.field_of_view = "toto"
    config.field_of_view = FieldOfView.FULL.value
    config.field_of_view = FieldOfView.HALF

    config_dict = config.to_dict()
    general_section_dict = config_dict["GENERAL_SECTION"]
    assert general_section_dict["input_folder"] == "my_folder"
    assert general_section_dict["output_file"] == "my_nx.nx"
    assert general_section_dict["file_extension"] == ".nx"
    assert general_section_dict["overwrite"] is True
    assert general_section_dict["dataset_basename"] == "test_"
    assert general_section_dict["dataset_info_file"] == "my_info_file.info"
    assert general_section_dict["title"] == "my title"
    assert general_section_dict["patterns_to_ignores"] == ("toto",)

    edf_headers_section_dict = config_dict["EDF_KEYS_SECTION"]
    assert (
        edf_headers_section_dict["motor_position_keys"] == settings.Tomo.EDF.MOTOR_POS
    )
    assert edf_headers_section_dict["motor_mne_keys"] == settings.Tomo.EDF.MOTOR_MNE
    assert edf_headers_section_dict["x_translation_keys"] == settings.Tomo.EDF.X_TRANS
    assert edf_headers_section_dict["y_translation_keys"] == settings.Tomo.EDF.Y_TRANS
    assert edf_headers_section_dict["z_translation_keys"] == settings.Tomo.EDF.Z_TRANS
    assert (
        edf_headers_section_dict["rotation_angle_keys"] == settings.Tomo.EDF.ROT_ANGLE
    )

    dark_flat_section_dict = config_dict["DARK_AND_FLAT_SECTION"]
    assert dark_flat_section_dict["dark_names_prefix"] == settings.Tomo.EDF.DARK_NAMES
    assert dark_flat_section_dict["flat_names_prefix"] == settings.Tomo.EDF.REFS_NAMES

    units_section_dict = config_dict["UNIT_SECTION"]
    assert units_section_dict["pixel_size_unit"] == str(_ureg.micrometer)
    assert units_section_dict["sample_detector_distance_unit"] == str(_ureg.meter)
    assert units_section_dict["energy_unit"] == str(_ureg.keV)
    assert units_section_dict["x_translation_unit"] == str(_ureg.meter)
    assert units_section_dict["y_translation_unit"] == str(_ureg.meter)
    assert units_section_dict["z_translation_unit"] == str(_ureg.meter)

    sample_section_dict = config_dict["SAMPLE_SECTION"]
    assert sample_section_dict["sample_name"] == "my sample"
    assert sample_section_dict["force_angle_calculation"] is True
    assert sample_section_dict["angle_calculation_endpoint"] is True
    assert sample_section_dict["angle_calculation_rev_neg_scan_range"] is False

    source_section_dict = config_dict["SOURCE_SECTION"]
    assert source_section_dict["instrument_name"] == "BMXX"
    assert source_section_dict["source_name"] == "XFEL"
    assert source_section_dict["source_type"] == SourceType.FIXED_TUBE_X_RAY.value

    detector_section_dict = config_dict["DETECTOR_SECTION"]
    assert detector_section_dict["field_of_view"] == FieldOfView.HALF.value



from __future__ import annotations

import os
import h5py
import numpy
import logging
from tqdm import tqdm

from silx.io.utils import open as open_hdf5
from silx.io.utils import h5py_read_dataset
from silx.io.url import DataUrl

from nxtomomill.utils.utils import str_datetime_to_numpy_datetime64
from nxtomomill.models.h52nx.FrameGroup import (
    FrameGroup,
    filter_acqui_frame_type,
)
from nxtomomill.models.h52nx._acquisitionstep import AcquisitionStep
from nxtomomill.io.config import TomoHDF5Config
from nxtomomill.models.utils import _get_title_dataset

from nxtomomill.utils.hdf5 import EntryReader
from nxtomomill.converter.hdf5.acquisition.multitomo import MultiTomoAcquisition
from nxtomomill.converter.hdf5.acquisition.backandforth import BackAndForthAcquisition

from .acquisition.utils.bliss_scan_type import get_bliss_scan_type
from .acquisition.baseacquisition import BaseAcquisition
from .acquisition.standardacquisition import StandardAcquisition
from .acquisition.zseriesacquisition import (
    ZSeriesBaseAcquisition,
    is_z_series_sequence,
    is_z_series_frm_translation_z,
)
from nxtomomill.utils._os import abspath

try:
    import hdf5plugin  # noqa F401
except ImportError:
    pass

_logger = logging.getLogger(__name__)


def is_multitomo_sequence(
    entry: h5py.Group, configuration: TomoHDF5Config, fallback_on_title: bool = True
) -> bool:
    """
    Check if the provided h5py.Group must be consider as an "initialization" of a sequence. And if so then if it is a multi-tomo sequence.
    It will first check for value contained in technique/scan_category else for the title name (legacy)
    """
    # check 'technique/scan_category' first
    scan_category = entry.get("technique/scan_category", None)
    if scan_category is not None:
        category_name = h5py_read_dataset(scan_category)
    elif fallback_on_title:
        # else fallback on title check
        category_name = _get_title_dataset(
            entry=entry, title_paths=BaseAcquisition.TITLE_PATHS
        )
    else:
        category_name = None

    if category_name is None:
        return False

    for multitomo_init_title in configuration.multitomo_init_titles:
        if category_name.startswith(multitomo_init_title):
            return True
    return False


def is_back_and_forth_sequence(
    entry: h5py.Group, configuration: TomoHDF5Config, fallback_on_title: bool = True
) -> bool:
    """
    Check if the provided h5py.Group must be consider as an "initialization" of a sequence. And if so then if it is a back-and-forth sequence.
    It will first check for value contained in technique/scan_category else for the title name (legacy)
    """
    # check 'technique/scan_category' first
    scan_category = entry.get("technique/scan_category", None)
    if scan_category is not None:
        category_name = h5py_read_dataset(scan_category)
    elif fallback_on_title:
        # else fallback on title check
        category_name = _get_title_dataset(
            entry=entry, title_paths=BaseAcquisition.TITLE_PATHS
        )
    else:
        category_name = None

    if category_name is None:
        return False

    for back_and_forth_init_title in configuration.back_and_forth_init_titles:
        if category_name.startswith(back_and_forth_init_title):
            return True
    return False


class _AcquisitionConstructorBase:
    """
    Base class of an acquisition constructor.
    The role of an acquisition constructor is to create instances of nxtomomill 'BaseAcquisition'.
    Then associate each bliss scan to his role (dark, flat, projs) and his (nxtomomill) acquisition.
    """

    def __init__(
        self,
        configuration: TomoHDF5Config,
        progress: tqdm | None,
        detector_sel_callback,
    ):
        """Constructor that build nxtomomill acquisition sequence from a set of entry"""
        self.configuration = configuration
        self.progress = progress
        self.detector_sel_callback = detector_sel_callback
        self._acquisitions = []

        self._current_acquisition = None
        """Pointer to the `BaseAcquisition` instance to be populated (we expect to add darks, flat and projections to it when iterating on bliss scans)"""
        self._start_index = 0
        """Counter on the index of the NXtomo to be created. This allow us to know from which nxtomo index a new `BaseAcquisition` must start from"""
        self._need_to_split_to_several_nxtomo = False
        """
        Flag required by multitomo and back-and-forth sequences.
        We need to wait until all bliss scans are registered to the nxtomomill Acquisition to update 'start_index'.
        It true at the start of a new the sequence we increase the start index by (nb_loop * nb_tomo) == expected of NXtomo generated by the previous sequence
        """

    def clear_acquisitions(self):
        self._acquisitions.clear()

    def build_sequence(self) -> tuple[BaseAcquisition]:
        raise NotImplementedError("Base class")

    @staticmethod
    def sort_fct(node_name: str, h5d: h5py.File):
        """
        sort the scan according to the 'start_time parameter'. If fails keep the original order.
        If a node has the 'is_rearranged' attribute then skip sort and keep the original sequence.
        """
        #
        node_link_to_treat = h5d.get(node_name, getlink=True)

        note_to_treat = h5d.get(node_name)
        is_rearranged = note_to_treat is not None and note_to_treat.attrs.get(
            "is_rearranged", False
        )
        # in some case the user might want to keep the order of the original sequence.
        # in this case we expect some preprocessing to be done and which has tag the node with the 'is_rearranged' attribute
        if is_rearranged:
            return False
        else:
            node = h5d.get(node_name)
            if node is not None:
                # node can be None in the case of a broken link
                start_time = node.get("start_time", None)
            else:
                _logger.warning(f"Broken link at {node_name}")
                start_time = None

            if start_time is not None:
                start_time = h5py_read_dataset(start_time)
                return str_datetime_to_numpy_datetime64(start_time)
            elif isinstance(node_link_to_treat, (h5py.ExternalLink, h5py.SoftLink)):
                return float(node_link_to_treat.path.split("/")[-1])
            else:
                # we expect to have node names like (1.1, 2.1...)
                return float(node_name)

    def build_progress(self, n: int) -> tqdm | None:
        """create tqdm progress bar"""
        if self.progress is not None:
            progress_read = tqdm(desc="read sequences")
            progress_read.total = n
            return progress_read

    def _update_start_index_for_acquisitions_splitting_nxtomos(
        self, url: DataUrl
    ) -> None:
        """
        Util called when processing `BaseAcquisition` instance splitting the NXtomo generated to update the '_start_index'
        This is the case for multi-tomo and back-and-forth tomo
        """
        nb_loop = self._current_acquisition.get_nb_loop(url)
        nb_tomo = self._current_acquisition.get_nb_tomo(url)
        if nb_loop is not None and nb_tomo is not None:
            self._start_index += int(nb_loop) * int(nb_tomo)
            self._need_to_split_to_several_nxtomo = False

    def _update_start_index_for_zseries(self, url: DataUrl) -> None:
        """
        Util called when processing `ZSeriesBaseAcquisition` instance splitting the NXtomo generated to update the '_start_index'
        """
        with EntryReader(url) as entry:
            z = self._current_acquisition.get_z(entry)
            if z not in self._acquisitions:
                self._start_index += 1


class _AcquisitionConstructorFromUrls(_AcquisitionConstructorBase):
    """
    Create (nxtomomill) acquisitions from a set of urls and roles (dark, flat, proj) provided by the users
    """

    def build_sequence(self) -> tuple[BaseAcquisition]:
        """
        Build acquisitions classes from the url definition

        :return:
        """
        self.clear_acquisitions()
        # when building from urls `tomo_n` has no meaning
        if self.configuration.check_tomo_n is None:
            self.configuration.check_tomo_n = False
        if self.configuration.is_using_titles:
            raise ValueError("Configuration specify that titles should be used")
        assert self.configuration.output_file is not None, "output_file requested"
        data_scans = self.configuration.data_scans
        # step 0: copy some urls instead if needed
        # update copy parameter
        for frame_grp in data_scans:
            if frame_grp.copy is None:
                frame_grp.copy = self.configuration.default_data_copy

        # step 1: if there is no init FrameGroup create an empty one because
        # this is requested
        if len(data_scans) == 0:
            return
        elif data_scans[0].frame_type is not AcquisitionStep.INITIALIZATION:
            data_scans = [
                FrameGroup(frame_type=AcquisitionStep.INITIALIZATION, url=None),
            ]
            data_scans.extend(self.configuration.data_scans)
            self.configuration.data_scans = tuple(data_scans)

        # step 2: treat FrameGroups
        progress_read = self.build_progress(len(data_scans))
        for frame_grp in data_scans:
            self.treat_bliss_scan(frame_grp, progress_read=progress_read)
        return tuple(self._acquisitions)

    def treat_bliss_scan(self, frame_grp: FrameGroup, progress_read: tqdm | None):
        """
        Treat a bliss scan from an instance of `FrameGroup`.
        This instance contains information provided by the user from the configuration file (url, frame_type and copy)

        :param frame_grp: Group to treat with (url, frame type and copy)
        :param progress_read: tqdm progress bar
        """
        if progress_read is not None:
            progress_read.update()
        # handle frame_type == init
        if frame_grp.frame_type is AcquisitionStep.INITIALIZATION:
            # re-init the constructor
            if self._need_to_split_to_several_nxtomo is True:
                _logger.warning(
                    f"Fail to retrieve expected number of nxtomo for {self._current_acquisition}"
                )
            self._need_to_split_to_several_nxtomo = False

            self._current_acquisition = self._get_current_acquisition(
                frame_grp=frame_grp
            )
            if isinstance(
                self._current_acquisition,
                (MultiTomoAcquisition, BackAndForthAcquisition),
            ):
                self._start_index += 0
                self._need_to_split_to_several_nxtomo = True
            elif isinstance(self._current_acquisition, StandardAcquisition):
                self._start_index += self._current_acquisition.get_expected_nx_tomo()
                self._need_to_split_to_several_nxtomo = False
            elif isinstance(self._current_acquisition, ZSeriesBaseAcquisition):
                self._need_to_split_to_several_nxtomo = False
                self._start_index += 0
            self._acquisitions.append(self._current_acquisition)
        # handle frame_type != init
        else:
            if self._current_acquisition is None:
                raise RuntimeError(
                    "processing error. Was not able to find acquisition initialization scan. Is it a bliss-tomo acquisition ?"
                )
            self._current_acquisition.register_step(
                url=frame_grp.url,
                entry_type=frame_grp.frame_type,
                copy_frames=frame_grp.copy_data,
            )

            # in case of z we append an index according to if
            # is already registered or not
            if isinstance(self._current_acquisition, ZSeriesBaseAcquisition):
                self._update_start_index_for_zseries(url=frame_grp.url)

            if self._need_to_split_to_several_nxtomo:
                if frame_grp.frame_type is AcquisitionStep.PROJECTION:
                    self._update_start_index_for_acquisitions_splitting_nxtomos(
                        url=frame_grp.url
                    )

    def _get_current_acquisition(self, frame_grp: FrameGroup):
        """Deduce the acquisition type from the frame group"""
        acqui_projs_fg = filter_acqui_frame_type(
            init=frame_grp,
            sequences=self.configuration.data_scans,
            frame_type=AcquisitionStep.PROJECTION,
        )
        acqui_projs_urls = tuple([acqui_proj.url for acqui_proj in acqui_projs_fg])
        h5f = None
        try:
            # open the initialization url
            if frame_grp.url is not None:
                h5f = h5py.File(frame_grp.url.file_path(), mode="r")
                init_entry = h5f.get(frame_grp.url.data_path())
            else:
                init_entry = None

            if is_z_series_frm_translation_z(acqui_projs_urls, self.configuration):
                return ZSeriesBaseAcquisition(
                    root_url=frame_grp.url,
                    configuration=self.configuration,
                    detector_sel_callback=self.detector_sel_callback,
                    start_index=self._start_index,
                )
            elif (init_entry is not None) and is_multitomo_sequence(
                init_entry, self.configuration
            ):
                return MultiTomoAcquisition(
                    root_url=frame_grp.url,
                    configuration=self.configuration,
                    detector_sel_callback=self.detector_sel_callback,
                    start_index=self._start_index,
                )
            elif (init_entry is not None) and is_back_and_forth_sequence(
                init_entry, self.configuration
            ):
                return BackAndForthAcquisition(
                    root_url=frame_grp.url,
                    configuration=self.configuration,
                    detector_sel_callback=self.detector_sel_callback,
                    start_index=self._start_index,
                )
            else:
                return StandardAcquisition(
                    root_url=frame_grp.url,
                    configuration=self.configuration,
                    detector_sel_callback=self.detector_sel_callback,
                    start_index=self._start_index,
                )
        finally:
            init_entry = None
            if h5f is not None:
                h5f.close()


class _AcquisitionConstructorFromTitles(_AcquisitionConstructorBase):
    """
    determine bliss scan role (dark, flat, proj, init) from titles.
    Create and populate instance of BaseAcquisition according to those roles.
    """

    def build_sequence(self) -> tuple[BaseAcquisition]:
        self.clear_acquisitions()
        if self.configuration.check_tomo_n is None:
            self.configuration.check_tomo_n = True

        with open_hdf5(self.configuration.input_file) as h5d:

            groups = list(h5d.keys())
            try:

                def sort(name):
                    return self.sort_fct(name, h5d=h5d)

                groups.sort(key=sort)
            except numpy.core._exceptions._UFuncNoLoopError:
                raise ValueError(
                    "Fail to order according to 'start_time'. Probably not all scans have a 'start_time' dataset"
                )

            # step 1: pre processing: group scan together
            progress_read = self.build_progress(len(groups))
            for group_name in groups:
                self.treat_bliss_scan(
                    group_name=group_name, h5d=h5d, progress_read=progress_read
                )
        return self._acquisitions

    def treat_bliss_scan(self, group_name: h5py.Group, h5d, progress_read: tqdm | None):
        """
        Treat a bliss scan (given as a group). Either create a new 'current acquisition' or register it to the previous (except if the scan is asked to be ignored by the use)

        :param group_name: group to be treat
        :param scan_group: h5d - h5py.File (open) containing the group.
        :param progress_read: optional progress to provide feedback
        """
        _logger.debug(f"parse {group_name}")
        if progress_read is not None:
            progress_read.update()

        try:
            entry = h5d[group_name]
        except KeyError:
            # case the key doesn't exist. Usual use case is that a bliss scan has been canceled
            _logger.warning(
                f"Unable to open {group_name} from {h5d.name}. Did the scan was canceled ? (Most likely). Skip this entry"
            )
            return

        # improve handling of External (this is the case of proposal files)
        if isinstance(h5d.get(group_name, getlink=True), h5py.ExternalLink):
            external_link = h5d.get(group_name, getlink=True)
            file_path = external_link.filename
            data_path = external_link.path
            if not os.path.isabs(file_path):
                file_path = abspath(
                    os.path.join(
                        os.path.dirname(self.configuration.input_file),
                        file_path,
                    )
                )
        else:
            file_path = self.configuration.input_file
            data_path = entry.name

        url = DataUrl(
            file_path=file_path,
            data_path=data_path,
            scheme="silx",
            data_slice=None,
        )

        entry_type = get_bliss_scan_type(url=url, configuration=self.configuration)
        if entry_type is AcquisitionStep.INITIALIZATION:
            if self._need_to_split_to_several_nxtomo is True:
                _logger.warning(
                    f"Fail to retrieve expected number of nxtomo for {self._current_acquisition}"
                )
            try:
                if is_z_series_sequence(entry=entry, configuration=self.configuration):
                    self._current_acquisition = ZSeriesBaseAcquisition(
                        root_url=url,
                        configuration=self.configuration,
                        detector_sel_callback=self.detector_sel_callback,
                        start_index=self._start_index,
                    )
                    self._start_index += (
                        self._current_acquisition.get_expected_nx_tomo()
                    )
                elif is_multitomo_sequence(
                    entry=entry, configuration=self.configuration
                ):
                    self._current_acquisition = MultiTomoAcquisition(
                        root_url=url,
                        configuration=self.configuration,
                        detector_sel_callback=self.detector_sel_callback,
                        start_index=self._start_index,
                    )
                    self._start_index += 0
                    self._need_to_split_to_several_nxtomo = True
                elif is_back_and_forth_sequence(
                    entry=entry, configuration=self.configuration
                ):
                    self._current_acquisition = BackAndForthAcquisition(
                        root_url=url,
                        configuration=self.configuration,
                        detector_sel_callback=self.detector_sel_callback,
                        start_index=self._start_index,
                    )
                    self._start_index += 0
                    self._need_to_split_to_several_nxtomo = True
                else:
                    self._current_acquisition = StandardAcquisition(
                        root_url=url,
                        configuration=self.configuration,
                        detector_sel_callback=self.detector_sel_callback,
                        start_index=self._start_index,
                    )
                    self._start_index += (
                        self._current_acquisition.get_expected_nx_tomo()
                    )
            except Exception as e:
                if self._ignore_entry(group_name):
                    return
                else:
                    raise e
            if self._ignore_entry(group_name):
                self._current_acquisition = None
                return

            self._acquisitions.append(self._current_acquisition)
        # continue "standard" tomo dataset handling
        elif self._current_acquisition is not None and not self._ignore_sub_entry(url):
            self._current_acquisition.register_step(
                url=url,
                entry_type=entry_type,
                copy_frames=self.configuration.default_data_copy,
            )
            # in case of z we append an index according to if
            # is already registered or not
            if isinstance(self._current_acquisition, ZSeriesBaseAcquisition):
                self._update_start_index_for_zseries(url=url)

            if self._need_to_split_to_several_nxtomo:
                if entry_type is AcquisitionStep.PROJECTION:
                    self._update_start_index_for_acquisitions_splitting_nxtomos(url=url)

        else:
            _logger.info(f"ignore entry {entry}")

    def _ignore_entry(self, group_name) -> bool:
        """check if the entry is part of the configuration entries (if provided)"""
        if len(self.configuration.entries) == 0:
            return False
        else:
            if not group_name.startswith("/"):
                group_name = "/" + group_name
            for entry in self.configuration.entries:
                if group_name == entry.data_path():
                    return False
            return True

    def _ignore_sub_entry(self, sub_entry_url: DataUrl | None) -> bool:
        """
        :return: True if the provided sub_entry should be ignored
        """
        if sub_entry_url is None:
            return False
        if not isinstance(sub_entry_url, DataUrl):
            raise TypeError(
                f"sub_entry_url is expected to be a DataUrl not {type(sub_entry_url)}"
            )
        if self.configuration.sub_entries_to_ignore is None:
            return False

        sub_entry_fp = sub_entry_url.file_path()
        sub_entry_dp = sub_entry_url.data_path()
        for entry in self.configuration.sub_entries_to_ignore:
            assert isinstance(entry, DataUrl)
            if entry.file_path() == sub_entry_fp and entry.data_path() == sub_entry_dp:
                return True
        return False



# coding: utf-8

"""
Application to convert a fluo-tomo dataset, after PyMCA (https://www.silx.org/doc/PyMca/dev/index.html) fit, into an hdf5/nexus file.

.. program-output:: nxtomomill fluo2nx --help

"""

import argparse
import logging

from nxtomomill import converter
from nxtomomill.models.blissfluo2nx import BlissFluo2nxModel

from tqdm import tqdm

logging.basicConfig(level=logging.INFO)


def main(argv):
    """ """
    parser = argparse.ArgumentParser(
        description="Converts Bliss fluo-tomo data (after PyMca fit) "
        "to hdf5 - nexus compliant file format."
    )
    parser.add_argument(
        "ewoksfluo_filename",
        help="Path to the ewoksfluo-generated (h5) filename taht contains fitted XRF data.",
    )
    parser.add_argument(
        "output_file",
        help="File produced by the converter. '.nx' extension recommended.",
    )
    parser.add_argument(
        "--detectors",
        nargs="*",
        default=list(),
        help="Define a list of (real or virtual) detector names used for the exp (space separated values - no comma). E.g. 'falcon xmap'. If not specified, all detectors are processed.",
    )
    parser.add_argument(
        "--dimension",
        help="2 for 2D XRFCT, 3 for 3D XRFCT. Default is 3.",
        default=3,
    )
    parser.add_argument(
        "--config",
        "--configuration-file",
        "--configuration",
        default=None,
        help="file containing the full configuration to convert from (PyMCA-computed) fluo projections to nexus. "
        "Default configuration file can be created from `nxtomomill fluo-config` command",
    )

    parser.add_argument(
        "--overwrite",
        default=False,
        action="store_true",
        help="If the output file exists then overwrite.",
    )

    options = parser.parse_args(argv[1:])
    config = BlissFluo2nxModel()
    if options.config is not None:
        config = config.from_cfg_file(options.config)

    check_input = {
        "ewoksfluo_filename": (
            options.ewoksfluo_filename,
            config.general_section.ewoksfluo_filename,
        ),
        "output file": (options.output_file, config.general_section.output_file),
    }
    for input_name, (opt_value, config_value) in check_input.items():
        if (
            opt_value is not None
            and config_value is not None
            and opt_value != config_value
        ):
            raise ValueError(
                f"two values for {input_name} are provided from arguments and configuration file ({opt_value, config_value})"
            )

    if options.ewoksfluo_filename is not None:
        config.general_section.ewoksfluo_filename = options.ewoksfluo_filename
    if options.output_file is not None:
        config.general_section.output_file = options.output_file
    if options.detectors is not None:
        config.general_section.detector_names = options.detectors
    config.general_section.dimension = options.dimension
    config.general_section.overwrite = options.overwrite

    converter.from_blissfluo_to_nx(
        configuration=config,
        progress=tqdm("blissfluo2nx"),
    )



import logging
from posixpath import join
import numpy as np
import pint
from tomoscan.esrf.scan.h5utils import get_h5_value
from nxtomo.application.nxtomo import NXtomo
from nxtomo.nxobject.nxdetector import ImageKey
from tomoscan.esrf.scan.fscan import FscanDataset, list_datasets
from silx.utils.deprecation import deprecated

_logger = logging.getLogger(__name__)
_ureg = pint.get_application_registry()


__all__ = [
    "from_fscan_to_nx",
]


@deprecated(reason="Removed as unused", since_version="2.0")
def from_fscan_to_nx(
    fname,
    output_fname,
    detector_name="pcoedgehs",
    rotation_motor_name="hrrz",
    halftomo=False,
    ignore_last_n_projections=0,
    energy=None,
    distance=None,
):
    entries = list_datasets(fname)
    if len(entries) < 3:
        _logger.error(
            "Error: Expected at least three datasets, got only %d" % (len(entries))
        )
        return
    do_360 = False
    if len(entries) == 6:
        # 360 degrees scans are done in two parts: [1.1, 2.1, 3.1] and then [4.1, 5.1]
        do_360 = True

    projs = FscanDataset(fname, detector_name=detector_name, entry="1.1")
    flats = FscanDataset(fname, detector_name=detector_name, entry="2.1")
    darks = FscanDataset(fname, detector_name=detector_name, entry="3.1")
    if do_360:
        projs2 = FscanDataset(fname, detector_name=detector_name, entry="4.1")
        flats2 = FscanDataset(fname, detector_name=detector_name, entry="5.1")
        # Some datasets don't have the 6.1 entry.
        # Anyway nabu does not support several series of darks.
        # For now it's safer to ignore the second series of darks
        try:
            darks2 = FscanDataset(fname, detector_name=detector_name, entry="6.1")
        except ValueError:
            _logger.error("Could not find entry 6.1, proceeding")
            darks2 = None
        if darks2 is not None:
            _logger.warning(
                "Discarding entry 6.1 as nabu does not support several series of darks yet"
            )

    my_nxtomo = NXtomo()

    data_urls = [darks.dataset_hdf5_url, flats.dataset_hdf5_url, projs.dataset_hdf5_url]
    if do_360:
        data_urls.extend(
            [
                # darks2.dataset_hdf5_url,
                flats2.dataset_hdf5_url,
                projs2.dataset_hdf5_url,
            ]
        )
    my_nxtomo.instrument.detector.data = data_urls

    img_keys = [
        [ImageKey.DARK_FIELD] * darks.data_shape[0],
        [ImageKey.FLAT_FIELD] * flats.data_shape[0],
        [ImageKey.PROJECTION] * (projs.data_shape[0] - ignore_last_n_projections),
        [ImageKey.ALIGNMENT] * ignore_last_n_projections,
    ]
    if do_360:
        img_keys.extend(
            [
                # [ImageKey.DARK_FIELD] * darks2.data_shape[0],
                [ImageKey.FLAT_FIELD] * flats2.data_shape[0],
                [ImageKey.PROJECTION]
                * (projs2.data_shape[0] - ignore_last_n_projections),
                [ImageKey.ALIGNMENT] * ignore_last_n_projections,
            ]
        )

    my_nxtomo.instrument.detector.image_key_control = np.concatenate(img_keys)

    rotation_angles_path = join(projs.entry, f"instrument/{rotation_motor_name}/data")
    rotation_angles = get_h5_value(projs.fname, rotation_angles_path)
    # Sometimes values are in /data, sometimes in /value ... who needs a stable format anyway ?
    if rotation_angles is None:
        _logger.error(
            f"Could not find the rotation angles in {rotation_angles_path}, trying in /values"
        )
    rotation_angles_path = rotation_angles_path.replace("/data", "/value")
    # ---
    rotation_angles = get_h5_value(projs.fname, rotation_angles_path)
    if rotation_angles is not None:
        last_idx = None
        if ignore_last_n_projections > 0:
            last_idx = -ignore_last_n_projections
        rotation_angles = [
            [0] * darks.data_shape[0],
            [0] * flats.data_shape[0],
            rotation_angles[:last_idx],
            [0] * ignore_last_n_projections,
        ]
        if do_360:
            rotation_angles2 = get_h5_value(
                projs2.fname,
                join(projs2.entry, f"instrument/{rotation_motor_name}/data"),
            )
            rotation_angles.extend(
                [
                    # [0] * darks2.data_shape[0],
                    [0] * flats2.data_shape[0],
                    rotation_angles2[:last_idx],
                    [0] * ignore_last_n_projections,
                ]
            )
        my_nxtomo.sample.rotation_angle = np.concatenate(rotation_angles) * _ureg.degree

    my_nxtomo.instrument.detector.field_of_view = (
        "Half" if (halftomo and do_360) else "Full"
    )
    my_nxtomo.instrument.detector.x_pixel_size = (
        my_nxtomo.instrument.detector.y_pixel_size
    ) = (6.5 * 1e-6)

    n_frames = len(my_nxtomo.sample.rotation_angle)
    my_nxtomo.instrument.detector.sequence_number = np.linspace(
        start=0, stop=n_frames, num=n_frames, dtype=np.uint32, endpoint=False
    )

    if energy is not None:
        my_nxtomo.energy = energy  # in keV by default
    if distance is not None:
        my_nxtomo.instrument.detector.distance = distance  # in meter

    my_nxtomo.save(
        file_path=output_fname,
        data_path="entry",
        overwrite=True,
        nexus_path_version=1.1,
    )



from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field, field_validator
from nxtomomill.models.utils import convert_str_to_tuple
from nxtomomill.settings import Tomo


class KeysSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    valid_camera_names: tuple[str, ...] = Field(
        default=tuple(), description="Valid camera names."
    )
    rotation_angle_keys: tuple[str, ...] = Field(
        default=Tomo.H5.ROT_ANGLE_KEYS,
        description="Keys for rotation angle.",
    )
    sample_x_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SAMPLE_X_KEYS,
        description="Keys for sample x translation.",
    )
    sample_y_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SAMPLE_Y_KEYS,
        description="Keys for sample y translation.",
    )
    translation_z_keys: tuple[str, ...] = Field(
        default=Tomo.H5.TRANSLATION_Z_KEYS,
        description="Keys for translation in z.",
    )
    translation_y_keys: tuple[str, ...] = Field(
        default=Tomo.H5.TRANSLATION_Y_KEYS,
        description="Keys for estimated center of rotation for half acquisition.",
    )
    diode_keys: tuple[str] = Field(
        default=Tomo.H5.DIODE_KEYS, description="Keys for diode."
    )
    exposure_time_keys: tuple[str, ...] = Field(
        default=Tomo.H5.ACQ_EXPO_TIME_KEYS, description="Keys for exposure time."
    )
    sample_x_pixel_size_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SAMPLE_X_PIXEL_SIZE_KEYS,
        description="Keys for sample x pixel size.",
    )
    sample_y_pixel_size_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SAMPLE_Y_PIXEL_SIZE_KEYS,
        description="Keys for sample y pixel size.",
    )
    detector_x_pixel_size_keys: tuple[str, ...] = Field(
        default=Tomo.H5.DETECTOR_X_PIXEL_SIZE_KEYS,
        description="Keys for detector x pixel size.",
    )
    detector_y_pixel_size_keys: tuple[str, ...] = Field(
        default=Tomo.H5.DETECTOR_Y_PIXEL_SIZE_KEYS,
        description="Keys for detector y pixel size.",
    )
    sample_detector_distance_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SAMPLE_DETECTOR_DISTANCE_KEYS,
        description="Keys for sample to detector distance.",
        alias="sample_detector_distance",
    )
    source_sample_distance_keys: tuple[str, ...] = Field(
        default=Tomo.H5.SOURCE_SAMPLE_DISTANCE_KEYS,
        description="Keys for source to sample distance.",
        alias="source_sample_distance",
    )
    machine_current_keys: tuple[str, ...] = Field(
        default=Tomo.H5.MACHINE_CURRENT_KEYS,
        description="Keys for machine current.",
    )

    @field_validator(
        "valid_camera_names",
        "rotation_angle_keys",
        "sample_x_keys",
        "sample_y_keys",
        "translation_z_keys",
        "translation_y_keys",
        "diode_keys",
        "exposure_time_keys",
        "sample_x_pixel_size_keys",
        "sample_y_pixel_size_keys",
        "detector_x_pixel_size_keys",
        "detector_y_pixel_size_keys",
        "sample_detector_distance_keys",
        "source_sample_distance_keys",
        "machine_current_keys",
        mode="plain",
    )
    @classmethod
    def cast_keys_section_to_tuple(cls, value: str) -> tuple[str, ...]:
        return convert_str_to_tuple(value) or ()



from __future__ import annotations

from nxtomomill.settings import Tomo
from nxtomomill.models.utils import convert_str_to_tuple

from pydantic import BaseModel, Field, field_validator, ConfigDict


class KeysSection(BaseModel):

    model_config = ConfigDict(validate_assignment=True, validate_by_name=True)

    motor_position_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.MOTOR_POS,
        description="Motor position key",
        alias="motor_position_key",
    )
    motor_mne_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.MOTOR_MNE,
        description="key to used for reading indices of each motor",
        alias="motor_mne_key",
    )
    rotation_angle_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.ROT_ANGLE,
        description="Keys to be used for reading rotation angle",
        alias="rot_angle_key",
    )
    x_translation_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.X_TRANS,
        description="Keys to be used for reading x translation",
        alias="x_translation_key",
    )
    y_translation_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.Y_TRANS,
        description="Keys to be used for reading y translation",
        alias="y_translation_key",
    )
    z_translation_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.Z_TRANS,
        description="Keys to be used for reading z translation",
        alias="z_translation_key",
    )
    machine_current_keys: tuple[str, ...] = Field(
        default=Tomo.EDF.MACHINE_CURRENT,
        description="Keys to be used for reading the machine current",
    )

    @field_validator(
        "motor_position_keys",
        "motor_mne_keys",
        "rotation_angle_keys",
        "x_translation_keys",
        "y_translation_keys",
        "z_translation_keys",
        "machine_current_keys",
        mode="plain",
    )
    @classmethod
    def cast_to_tuple(cls, value: str | tuple[str, ...]) -> tuple[str,]:
        if isinstance(value, tuple):
            return value
        return convert_str_to_tuple(value) or ()



# coding: utf-8

import os
import pytest

from nxtomomill.io.config.confighandler import TomoHDF5ConfigHandler
from nxtomomill.io.config.hdf5config import TomoHDF5Config
from nxtomomill.models.h52nx.FrameGroup import FrameGroup


class _ARParseMock(object):
    def __init__(self):
        self.config_file = None
        self.input_file = None
        self.output_file = None
        self.set_params = None
        self.alignment_titles = None
        self.projection_titles = None
        self.flat_titles = None
        self.dark_titles = None
        self.zseries_init_titles = None
        self.multitomo_init_titles = None
        self.init_titles = None
        self.back_and_forth_init_titles = None
        self.sample_x_pixel_size_keys = None
        self.sample_y_pixel_size_keys = None
        self.exposure_time_keys = None
        self.rotation_angle_keys = None
        self.valid_camera_names = None
        self.translation_z_keys = None
        self.sample_y_keys = None
        self.sample_x_keys = None
        self.request_input = False
        self.raises_error = False
        self.default_data_copy = None
        self.sub_entries_to_ignore = None
        self.entries = None
        self.debug = False
        self.overwrite = False
        self.single_file = False
        self.no_master_file = False
        self.file_extension = None
        self.field_of_view = None
        self.sample_detector_distance_keys = None


def test_creation_from_config_file(tmp_path):
    options = _ARParseMock()
    with pytest.raises(ValueError):
        TomoHDF5ConfigHandler(argparse_options=options)

    input_file_path = os.path.join(tmp_path, "output_file.cfg")
    input_config = TomoHDF5Config()
    input_config.to_cfg_file(input_file_path)

    options.config_file = input_file_path

    # insure this is valid if we add an input file and an output file
    # from the command line
    options.input_file = "toto.h5"
    options.output_file = "toto.nx"
    TomoHDF5ConfigHandler(argparse_options=options)

    # or if we provide them from the configuration file
    options.input_file = None
    options.output_file = None
    input_config.input_file = "toto.h5"
    input_config.output_file = "toto.nx"
    input_config.to_cfg_file(input_file_path)
    TomoHDF5ConfigHandler(argparse_options=options)


def write_and_read_configuration(folder, data_urls):
    configuration = TomoHDF5Config()
    configuration.data_scans = data_urls
    configuration.input_file = "toto.h5"
    configuration.output_file = "toto.nx"

    file_path = os.path.join(folder, "h52nx.cfg")
    configuration.to_cfg_file(file_path=file_path)

    options = _ARParseMock()
    options.config_file = file_path
    config_handler = TomoHDF5ConfigHandler(options)
    return config_handler.configuration


def test_local_paths(tmp_path):
    res_config = write_and_read_configuration(
        folder=tmp_path,
        data_urls=(
            FrameGroup(url="/entry0000/data/flats", frame_type="flat"),
            FrameGroup(url="/entry0000/data/projection1", frame_type="projection"),
            FrameGroup(url="/entry0000/data/projection2", frame_type="projection"),
            FrameGroup(url="/entry0000/data/alignment1", frame_type="alignment"),
            FrameGroup(url="/entry0000/data/alignment2", frame_type="alignment"),
            FrameGroup(url="/entry0000/data/darks", frame_type="dark"),
        ),
    )
    assert res_config is not None
    assert len(res_config.data_scans) == 6
    # check flats
    flat_url = res_config.data_scans[0].url
    assert flat_url.file_path() in (None, "")
    assert flat_url.data_path() == "/entry0000/data/flats"
    assert flat_url.data_slice() in (None, "")
    # check projections
    projection_url_0 = res_config.data_scans[1].url
    projection_url_1 = res_config.data_scans[2].url
    assert projection_url_0.file_path() in (None, "")
    assert projection_url_0.data_path() == "/entry0000/data/projection1"
    assert projection_url_0.data_slice() is None
    assert projection_url_1.file_path() in (None, "")
    assert projection_url_1.data_path() == "/entry0000/data/projection2"
    assert projection_url_1.data_slice() is None

    # check darks
    dark_url = res_config.data_scans[-1].url
    assert dark_url.file_path() in (None, "")
    assert dark_url.data_path(), "/entry0000/data/darks"
    assert dark_url.data_slice() is None

    # check alignments
    alignment_url_0 = res_config.data_scans[3].url
    alignment_url_1 = res_config.data_scans[4].url
    assert alignment_url_0.file_path() in (None, "")
    assert alignment_url_0.data_path(), "/entry0000/data/alignment1"
    assert alignment_url_0.data_slice() is None
    assert alignment_url_1.data_path() == "/entry0000/data/alignment2"


def test_external_paths(tmp_path):
    """
    Check that DataUrl are handled.

    Warning: this also check data_slices but those are not handled.
    """
    res_config = write_and_read_configuration(
        folder=tmp_path,
        data_urls=(
            FrameGroup(
                frame_type="flat",
                url="silx:///myfile.h5?path=/entry0000/data/flats",
            ),
            FrameGroup(
                frame_type="dark",
                url="h5py:///data/file2.hdf5?path=/entry0000/data/darks",
            ),
            FrameGroup(frame_type="proj", url="h5py:///data/file3.hdf5?path=/data"),
            FrameGroup(
                frame_type="proj",
                url="silx:///data/file2.hdf5?path=/entry0000/data/projection2&slice=5:100",
            ),
            FrameGroup(
                frame_type="alignment",
                url="silx:///myfile.h5?path=/entry0000/data/alignment&slice=5",
            ),
        ),
    )
    assert res_config is not None
    assert len(res_config.data_scans) == 5
    # check flats
    flat_url = res_config.data_scans[0]
    assert flat_url.url.file_path() == "/myfile.h5"
    assert flat_url.url.data_path() == "/entry0000/data/flats"
    assert flat_url.url.data_slice() is None
    assert flat_url.url.scheme() == "silx"

    # check projections
    projection_url_0 = res_config.data_scans[2].url
    projection_url_1 = res_config.data_scans[3].url
    assert projection_url_0.file_path() == "/data/file3.hdf5"
    assert projection_url_0.data_path() == "/data"
    assert projection_url_0.data_slice() is None
    assert projection_url_0.scheme() == "h5py"
    assert projection_url_1.file_path() == "/data/file2.hdf5"
    assert projection_url_1.data_path() == "/entry0000/data/projection2"
    assert projection_url_1.data_slice() == (slice(5, 100),)
    assert projection_url_1.scheme() == "silx"

    # check darks
    dark_url = res_config.data_scans[1].url
    assert dark_url.file_path() == "/data/file2.hdf5"
    assert dark_url.data_path() == "/entry0000/data/darks"
    assert dark_url.data_slice() is None
    assert dark_url.scheme() == "h5py"

    # check alignments
    alignment_url_0 = res_config.data_scans[4].url
    assert alignment_url_0.file_path() == "/myfile.h5"
    assert alignment_url_0.data_path() == "/entry0000/data/alignment"
    assert alignment_url_0.data_slice() == (5,)
    assert alignment_url_0.scheme() == "silx"



# coding: utf-8

import logging

import pytest
import numpy

from nxtomomill.converter.fluo.fluoscan import FluoTomoScan2D
from nxtomomill.tests.datasets import GitlabDataset

logging.disable(logging.INFO)


@pytest.fixture(scope="class")
def fluodata2D(request):
    cls = request.cls
    cls.scan_dir = GitlabDataset.get_dataset("fluo_datasets2D")
    cls.dataset_basename = "CONT2_p2_600nm_FT02_slice_0"
    cls.scan = FluoTomoScan2D(
        scan=cls.scan_dir,
        dataset_basename=cls.dataset_basename,
        detectors=(),
    )


@pytest.mark.usefixtures("fluodata2D")
class TestFluo2D:
    def test_all_detectors(self):
        assert (
            len(self.scan.el_lines) == 2  # pylint: disable=E1101
        ), f"Number of elements found should be 2 and is {len(self.scan.el_lines)}."  # pylint: disable=E1101
        assert set(self.scan.detectors) == set(  # pylint: disable=E1101
            ["fluo1", "corrweighted"]
        ), f"There should be 2 'detectors' (fluo1 and corrweighted), {len(self.detectors)} were found."  # pylint: disable=E1101

    def test_one_detector(self):
        scan = FluoTomoScan2D(
            scan=self.scan_dir,  # pylint: disable=E1101
            dataset_basename=self.dataset_basename,  # pylint: disable=E1101
            detectors=("corrweighted",),
        )
        assert (
            len(scan.el_lines) == 2
        ), f"Number of elements found should be 2 and is {len(scan.el_lines)}."
        assert (
            len(scan.detectors) == 1
        ), f"There should be 1 detector (corrweighted), {len(scan.detectors)} were found."

        # One ghost detector (no corresponding files)
        # test general section setters
        with pytest.raises(ValueError):
            scan = FluoTomoScan2D(
                scan=self.scan_dir,  # pylint: disable=E1101
                dataset_basename=self.dataset_basename,  # pylint: disable=E1101
                detectors=("toto",),
            )

    def test_load_data(self):
        data = self.scan.load_data("corrweighted", "Ca", 0)  # pylint: disable=E1101
        assert data.shape == (1, 251, 1000)

    def test_load_energy_and_pixel_size(self):
        assert self.scan.energy == 17.1  # pylint: disable=E1101
        assert numpy.allclose(
            self.scan.pixel_size, 6e-10, atol=1e-4  # pylint: disable=E1101
        )  # Tolerance:0.1nm (since pixel_size is expected in um).



from .detector import has_valid_detector  # noqa F401
from .detector import get_nx_detectors  # noqa F401
from .detector import guess_nx_detector  # noqa F401
from .bliss_scan_type import get_bliss_scan_type  # noqa F401
from .bliss_scan_type import get_entry_type  # noqa F401
from .machine_current import deduce_machine_current  # noqa F401
from .timestamps import split_timestamps  # noqa F401
from .series import group_series  # noqa F401
